import numpy as np
import os

cosmo1 = {'name': 'cosmo1', 'Omega_m': 0.3, 'Omega_b': 0.05, 'h': 0.67, 'n_s': 0.96, 'sigma8': 0.8, 'w0': -1.0, 'wa': 0.0, 'omega_nu': 0.0}
cosmo2 = {'name': 'cosmo2', 'Omega_m': 0.3, 'Omega_b': 0.05, 'h': 0.67, 'n_s': 0.96, 'sigma8': 0.8, 'w0': -1.0, 'wa': 0.0, 'omega_nu': 0.001}

def get_bm(cosmo, kind):
    cosmoname = cosmo['name']
    f = open("xstar.dat", "w")
    h = cosmo['h']
    ns = cosmo['n_s']
    s8 = cosmo['sigma8']
    w0 = cosmo['w0']
    wa = cosmo['wa']
    om = cosmo['Omega_m']*h**2
    ob = cosmo['Omega_b']*h**2
    onu = cosmo['omega_nu']
    f.write(f'{om} {ob} {s8} {h} {ns} {w0} {wa} {onu} 0.0\n')
    f.write(f'{om} {ob} {s8} {h} {ns} {w0} {wa} {onu} 1.0\n')
    f.close()
    os.system(f'./P_{kind}/emu.exe')
    for i in range(2):
        os.system(f'mv EMU{i}.txt ../../{cosmoname}_MTII_{kind}_{i}.txt')

get_bm(cosmo1, 'tot')
get_bm(cosmo1, 'cb')
get_bm(cosmo2, 'tot')
get_bm(cosmo2, 'cb')



__all__ = ("HaloProfileCIBShang12",)

import numpy as np
from scipy.integrate import simpson
from scipy.special import lambertw

from . import HaloProfileNFW, HaloProfileCIB


class HaloProfileCIBShang12(HaloProfileCIB):
    """ CIB profile implementing the model by `Shang et al. 2012
    <https://arxiv.org/abs/1109.1522>`_.

    The parametrization for the mean profile emissivity :math:`j_\\nu`
    is:

    .. math::
        j_\\nu(r) = \\frac{1}{4\\pi}
        \\left(L^{\\rm cen}_{\\nu(1+z)}(M)+
        L^{\\rm sat}_{\\nu(1+z)}u_{\\rm sat}(r|M)
        \\right),

    where the luminosity from centrals and satellites is
    modelled as:

    .. math::
        L^{\\rm cen}_{\\nu}(M) = L^{\\rm gal}_\\nu(M)\\,
        N_{\\rm cen}(M),

    .. math::
        L^{\\rm sat}_{\\nu}(M) = \\int_{M_{\\rm min}}^{M} dm
        \\frac{dN_{\\rm sub}}{dm}\\,L^{\\rm gal}_\\nu(m).

    Here, :math:`dN_{\\rm sub}/dm` is the subhalo mass function,
    :math:`u_{\\rm sat}` is the satellite galaxy density profile
    (modelled as a truncated NFW profile), and the infrared
    galaxy luminosity is parametrized as

    .. math::
        L^{\\rm gal}_{\\nu}(M,z)=L_0(1+z)^{s_z}\\,
        \\Sigma(M)\\,S_\\nu,

    where the mass dependence is lognormal

    .. math::
        \\Sigma(M) = \\frac{M}{\\sqrt{2\\pi\\sigma_{LM}^2}}
        \\exp\\left[-\\frac{\\log_{10}^2(M/M_{\\rm eff})}
        {2\\sigma_{LM}^2}\\right],

    and the spectrum is a modified black-body law

    .. math::
        S_\\nu \\propto\\left\\{
        \\begin{array}{cc}
           \\nu^\\beta\\,B_\\nu(T_d) & \\nu < \\nu_0 \\\\
           \\nu^\\gamma & \\nu \\geq \\nu_0
        \\end{array}
        \\right.,

    with the normalization fixed by :math:`S_{\\nu_0}=1`,
    and :math:`\\nu_0` defined so the spectrum has a continuous
    derivative for all :math:`\\nu`.

    Finally, the dust temperature is assumed to have a redshift
    dependence of the form :math:`T_d=T_0(1+z)^\\alpha`.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        concentration (:class:`~pyccl.halos.halo_model_base.Concentration`):
            concentration-mass relation for NFW profile.
        nu_GHz (:obj:`float`): frequency in GHz.
        alpha (:obj:`float`): dust temperature evolution parameter.
        T0 (:obj:`float`): dust temperature at :math:`z=0` in Kelvin.
        beta (:obj:`float`): dust spectral index.
        gamma (:obj:`float`): high frequency slope.
        s_z (:obj:`float`): luminosity evolution slope.
        log10Meff (:obj:`float`): :math:`\\log_{10}` of the most efficient mass.
        siglog10M (:obj:`float`): logarithmic scatter in mass.
        Mmin (:obj:`float`): minimum subhalo mass.
        L0 (:obj:`float`): luminosity scale (in
            :math:`{\\rm Jy}\\,{\\rm Mpc}^2\\,M_\\odot^{-1}`).
    """ # noqa
    __repr_attrs__ = __eq_attrs__ = (
        "nu", "alpha", "T0", "beta", "gamma", "s_z", "log10Meff", "siglog10M",
        "Mmin", "L0", "mass_def", "concentration", "precision_fftlog",)
    _one_over_4pi = 0.07957747154

    def __init__(self, *, mass_def, concentration, nu_GHz, alpha=0.36,
                 T0=24.4, beta=1.75, gamma=1.7, s_z=3.6, log10Meff=12.6,
                 siglog10M=0.707, Mmin=1E10, L0=6.4E-8):
        self.nu = nu_GHz
        self.alpha = alpha
        self.T0 = T0
        self.beta = beta
        self.gamma = gamma
        self.s_z = s_z
        self.log10Meff = log10Meff
        self.siglog10M = siglog10M
        self.Mmin = Mmin
        self.L0 = L0
        kwargs = {"concentration": concentration, "mass_def": mass_def}
        self.pNFW = HaloProfileNFW(**kwargs)
        super().__init__(**kwargs)

    def dNsub_dlnM_TinkerWetzel10(self, Msub, Mparent):
        """Subhalo mass function of `Tinker & Wetzel 2010
        <https://arxiv.org/abs/0909.1325>`_. Number of subhalos
        per (natural) logarithmic interval of mass.

        Args:
            Msub (:obj:`float` or `array`): sub-halo mass (in solar masses).
            Mparent (:obj:`float`): parent halo mass (in solar masses).

        Returns:
            (:obj:`float` or `array`): average number of subhalos.
        """
        return 0.30*(Msub/Mparent)**(-0.7)*np.exp(-9.9*(Msub/Mparent)**2.5)

    def update_parameters(self, nu_GHz=None,
                          alpha=None, T0=None, beta=None, gamma=None,
                          s_z=None, log10Meff=None, siglog10M=None,
                          Mmin=None, L0=None):
        """ Update any of the parameters associated with
        this profile. Any parameter set to ``None`` won't be updated.

        Args:
            nu_GHz (:obj:`float`): frequency in GHz.
            alpha (:obj:`float`): dust temperature evolution parameter.
            T0 (:obj:`float`): dust temperature at :math:`z=0` in Kelvin.
            beta (:obj:`float`): dust spectral index.
            gamma (:obj:`float`): high frequency slope.
            s_z (:obj:`float`): luminosity evolution slope.
            log10Meff (:obj:`float`): :math:`\\log_{10}` of the most
                efficient mass.
            siglog10M (:obj:`float`): logarithmic scatter in mass.
            Mmin (:obj:`float`): minimum subhalo mass.
            L0 (:obj:`float`): luminosity scale (in
                :math:`{\\rm Jy}\\,{\\rm Mpc}^2\\,M_\\odot^{-1}`).
        """
        if nu_GHz is not None:
            self.nu = nu_GHz
        if alpha is not None:
            self.alpha = alpha
        if T0 is not None:
            self.T0 = T0
        if beta is not None:
            self.beta = beta
        if gamma is not None:
            self.gamma = gamma
        if s_z is not None:
            self.s_z = s_z
        if log10Meff is not None:
            self.log10Meff = log10Meff
        if siglog10M is not None:
            self.siglog10M = siglog10M
        if Mmin is not None:
            self.Mmin = Mmin
        if L0 is not None:
            self.L0 = L0

    def _spectrum(self, nu, a):
        # h*nu_GHZ / k_B / Td_K
        h_GHz_o_kB_K = 0.0479924466
        Td = self.T0/a**self.alpha
        x = h_GHz_o_kB_K * nu / Td

        # Find nu_0
        q = self.beta+3+self.gamma
        x0 = q+np.real(lambertw(-q*np.exp(-q), k=0))

        def mBB(x):
            ex = np.exp(x)
            return x**(3+self.beta)/(ex-1)

        mBB0 = mBB(x0)

        def plaw(x):
            return mBB0*(x0/x)**self.gamma

        return np.piecewise(x, [x <= x0],
                            [mBB, plaw])/mBB0

    def _Lum(self, l10M, a):
        # Redshift evolution
        phi_z = a**(-self.s_z)
        # Mass dependence
        # M/sqrt(2*pi*siglog10M^2)
        sig_pref = 10**l10M/(2.50662827463*self.siglog10M)
        sigma_m = sig_pref * np.exp(-0.5*((l10M - self.log10Meff)
                                          / self.siglog10M)**2)
        return self.L0*phi_z*sigma_m

    def _Lumcen(self, M, a):
        Lum = self._Lum(np.log10(M), a)
        Lumcen = np.heaviside(M-self.Mmin, 1)*Lum
        return Lumcen

    def _Lumsat(self, M, a):
        if not np.max(M) > self.Mmin:
            return np.zeros_like(M)

        res = np.zeros_like(M)
        M_use = M[M >= self.Mmin, None]
        logM = np.log10(M_use)
        LOGM_MIN = np.log10(self.Mmin)
        nm = max(2, 10*int(np.max(logM) - LOGM_MIN))
        msub = np.linspace(LOGM_MIN, np.max(logM), nm+1)[None, :]

        Lum = self._Lum(msub, a)
        dnsubdlnm = self.dNsub_dlnM_TinkerWetzel10(10**msub, M_use)
        integ = dnsubdlnm * Lum
        Lumsat = simpson(integ, x=np.log(10)*msub)
        res[-len(Lumsat):] = Lumsat
        return res

    def _real(self, cosmo, r, M, a):
        M_use = np.atleast_1d(M)
        r_use = np.atleast_1d(r)

        # (redshifted) Frequency dependence
        spec_nu = self._spectrum(self.nu/a, a)

        Ls = self._Lumsat(M_use, a)
        ur = self.pNFW._real(cosmo, r_use, M_use, a)/M_use[:, None]

        prof = Ls[:, None]*ur*spec_nu*self._one_over_4pi

        if np.ndim(r) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def _fourier(self, cosmo, k, M, a):
        M_use = np.atleast_1d(M)
        k_use = np.atleast_1d(k)

        # (redshifted) Frequency dependence
        spec_nu = self._spectrum(self.nu/a, a)

        Lc = self._Lumcen(M_use, a)
        Ls = self._Lumsat(M_use, a)
        uk = self.pNFW._fourier(cosmo, k_use, M_use, a)/M_use[:, None]

        prof = (Lc[:, None]+Ls[:, None]*uk)*spec_nu*self._one_over_4pi

        if np.ndim(k) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def _fourier_variance(self, cosmo, k, M, a, nu_other=None):
        M_use = np.atleast_1d(M)
        k_use = np.atleast_1d(k)

        spec_nu1 = self._spectrum(self.nu/a, a)
        if nu_other is None:
            spec_nu2 = spec_nu1
        else:
            spec_nu2 = self._spectrum(nu_other/a, a)

        Lc = self._Lumcen(M_use, a)
        Ls = self._Lumsat(M_use, a)
        uk = self.pNFW._fourier(cosmo, k_use, M_use, a)/M_use[:, None]

        prof = Ls[:, None]*uk
        prof = 2*Lc[:, None]*prof + prof**2
        prof *= spec_nu1*spec_nu2*self._one_over_4pi**2

        if np.ndim(k) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof



import os
import numpy as np
import pyccl as ccl

# Set cosmology
cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, Omega_g=0, Omega_k=0,
                      h=0.7, sigma8=0.8, n_s=0.96, Neff=0, m_nu=0.0,
                      w0=-1, wa=0, T_CMB=2.7255,
                      transfer_function='eisenstein_hu')
# Read data
dirdat = os.path.join(os.path.dirname(__file__), 'data')

# Redshifts
zs = np.array([0., 0.5, 1.])


def test_hmf_despali16():
    hmd = ccl.halos.MassDef('vir', 'critical')
    mf = ccl.halos.MassFuncDespali16(mass_def=hmd)
    d_hmf = np.loadtxt(os.path.join(dirdat, 'hmf_despali16.txt'),
                       unpack=True)
    m = d_hmf[0]
    for iz, z in enumerate(zs):
        nm_d = d_hmf[iz+1]
        nm_h = mf(cosmo, m, 1. / (1 + z))
        assert np.all(np.fabs(nm_h / nm_d - 1) < 0.01)


def test_hmf_bocquet16():
    mf = ccl.halos.MassFuncBocquet16(mass_def="200c")
    d_hmf = np.loadtxt(os.path.join(dirdat, 'hmf_bocquet16.txt'),
                       unpack=True)
    m = d_hmf[0]
    for iz, z in enumerate(zs):
        nm_d = d_hmf[iz+1]
        nm_h = mf(cosmo, m, 1. / (1 + z))
        assert np.all(np.fabs(nm_h / nm_d - 1) < 0.01)


def test_hmf_watson13():
    mf = ccl.halos.MassFuncWatson13()
    d_hmf = np.loadtxt(os.path.join(dirdat, 'hmf_watson13.txt'),
                       unpack=True)
    m = d_hmf[0]
    for iz, z in enumerate(zs):
        nm_d = d_hmf[iz+1]
        nm_h = mf(cosmo, m, 1. / (1 + z))
        assert np.all(np.fabs(nm_h / nm_d - 1) < 0.01)


def test_hmf_tinker08():
    mf = ccl.halos.MassFuncTinker08()
    d_hmf = np.loadtxt(os.path.join(dirdat, 'hmf_tinker08.txt'),
                       unpack=True)
    m = d_hmf[0]
    for iz, z in enumerate(zs):
        nm_d = d_hmf[iz+1]
        nm_h = mf(cosmo, m, 1. / (1 + z))
        assert np.all(np.fabs(nm_h / nm_d - 1) < 0.01)


def test_hmf_press74():
    mf = ccl.halos.MassFuncPress74()
    d_hmf = np.loadtxt(os.path.join(dirdat, 'hmf_press74.txt'),
                       unpack=True)
    m = d_hmf[0]
    for iz, z in enumerate(zs):
        nm_d = d_hmf[iz+1]
        nm_h = mf(cosmo, m, 1. / (1 + z))
        assert np.all(np.fabs(nm_h / nm_d - 1) < 0.01)


def test_hmf_angulo12():
    mf = ccl.halos.MassFuncAngulo12()
    d_hmf = np.loadtxt(os.path.join(dirdat, 'hmf_angulo12.txt'),
                       unpack=True)
    m = d_hmf[0]
    for iz, z in enumerate(zs):
        nm_d = d_hmf[iz+1]
        nm_h = mf(cosmo, m, 1. / (1 + z))
        assert np.all(np.fabs(nm_h / nm_d - 1) < 0.01)


def test_hmf_sheth99():
    mf = ccl.halos.MassFuncSheth99()
    d_hmf = np.loadtxt(os.path.join(dirdat, 'hmf_sheth99.txt'),
                       unpack=True)
    m = d_hmf[0]
    for iz, z in enumerate(zs):
        nm_d = d_hmf[iz+1]
        nm_h = mf(cosmo, m, 1. / (1 + z))
        assert np.all(np.fabs(nm_h / nm_d - 1) < 0.01)


def test_hmf_jenkins01():
    mf = ccl.halos.MassFuncJenkins01()
    d_hmf = np.loadtxt(os.path.join(dirdat, 'hmf_jenkins01.txt'),
                       unpack=True)
    m = d_hmf[0]
    for iz, z in enumerate(zs):
        nm_d = d_hmf[iz+1]
        nm_h = mf(cosmo, m, 1. / (1 + z))
        assert np.all(np.fabs(nm_h / nm_d - 1) < 0.01)



__all__ = ("halomod_mean_profile_1pt", "halomod_bias_1pt",)

import numpy as np


def _Ix1(func, cosmo, hmc, k, a, prof):
    # I_X_1 dispatcher for internal use
    """
    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        hmc (:class:`~pyccl.halos.halo_model.HMCalculator`): a halo model
            calculator.
        k (:obj:`float` or `array`): comoving wavenumber in
            :math:`{\\rm Mpc}^{-1}`.
        a (:obj:`float` or `array`): scale factor.
        prof (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile.

    Returns:
        (:obj:`float` or `array`): integral values evaluated at each
        combination of ``k`` and ``a``. The shape of the output will
        be ``(N_a, N_k)`` where ``N_k`` and ``N_a`` are the sizes of
        ``k`` and ``a`` respectively. If ``k`` or ``a`` are scalars, the
        corresponding dimension will be squeezed out on output.
    """
    func = getattr(hmc, func)

    a_use = np.atleast_1d(a).astype(float)
    k_use = np.atleast_1d(k).astype(float)

    na = len(a_use)
    nk = len(k_use)
    out = np.zeros([na, nk])
    for ia, aa in enumerate(a_use):
        i11 = func(cosmo, k_use, aa, prof)
        norm = prof.get_normalization(cosmo, aa, hmc=hmc)
        out[ia] = i11 / norm

    if np.ndim(a) == 0:
        out = np.squeeze(out, axis=0)
    if np.ndim(k) == 0:
        out = np.squeeze(out, axis=-1)
    return out


def halomod_mean_profile_1pt(cosmo, hmc, k, a, prof):
    """ Returns the mass-weighted mean halo profile.

    .. math::
        I^0_1(k,a|u) = \\int dM\\,n(M,a)\\,\\langle u(k,a|M)\\rangle,

    where :math:`n(M,a)` is the halo mass function, and
    :math:`\\langle u(k,a|M)\\rangle` is the halo profile as a
    function of scale, scale factor and halo mass.
    """
    return _Ix1("I_0_1", cosmo, hmc, k, a, prof)


def halomod_bias_1pt(cosmo, hmc, k, a, prof):
    """ Returns the mass-and-bias-weighted mean halo profile.

    .. math::
        I^1_1(k,a|u) = \\int dM\\,n(M,a)\\,b(M,a)\\,
        \\langle u(k,a|M)\\rangle,

    where :math:`n(M,a)` is the halo mass function,
    :math:`b(M,a)` is the halo bias, and
    :math:`\\langle u(k,a|M)\\rangle` is the halo profile as a
    function of scale, scale factor and halo mass.
    """
    return _Ix1("I_1_1", cosmo, hmc, k, a, prof)


halomod_mean_profile_1pt.__doc__ += _Ix1.__doc__
halomod_bias_1pt.__doc__ += _Ix1.__doc__



import numpy as np
import pytest
import pyccl as ccl
from pyccl.modified_gravity import MuSigmaMG


ccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = False
COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')
PKA = ccl.Pk2D.from_function(pkfunc=lambda k, a: np.log(a/k))
ZZ = np.linspace(0., 1., 200)
NN = np.exp(-((ZZ-0.5)/0.1)**2)
LENS = ccl.WeakLensingTracer(COSMO, dndz=(ZZ, NN))


@pytest.mark.parametrize('p_of_k_a', [ccl.DEFAULT_POWER_SPECTRUM, PKA])
def test_cells_smoke(p_of_k_a):
    # make a set of tracers to test with
    z = np.linspace(0., 1., 200)
    n = np.exp(-((z-0.5)/0.1)**2)
    b = np.sqrt(1. + z)
    lens1 = ccl.WeakLensingTracer(COSMO, dndz=(z, n))
    lens2 = ccl.WeakLensingTracer(COSMO, dndz=(z, n), ia_bias=(z, n))
    nc1 = ccl.NumberCountsTracer(COSMO, has_rsd=False, dndz=(z, n),
                                 bias=(z, b))
    nc2 = ccl.NumberCountsTracer(COSMO, has_rsd=True, dndz=(z, n), bias=(z, b))
    nc3 = ccl.NumberCountsTracer(
        COSMO, has_rsd=True, dndz=(z, n), bias=(z, b), mag_bias=(z, b))
    cmbl = ccl.CMBLensingTracer(COSMO, z_source=1100.)
    tracers = [lens1, lens2, nc1, nc2, nc3, cmbl]

    ell_scl = 4.
    ell_int = 4
    ell_lst = [2, 3, 4, 5]
    ell_arr = np.arange(2, 5)
    ells = [ell_int, ell_scl, ell_lst, ell_arr]

    for i in range(len(tracers)):
        for j in range(i, len(tracers)):
            for ell in ells:
                corr = ccl.angular_cl(
                    COSMO, tracers[i], tracers[j], ell, p_of_k_a=p_of_k_a)
                assert np.all(np.isfinite(corr))
                assert np.shape(corr) == np.shape(ell)

                # reversing should be fine
                corr_rev = ccl.angular_cl(
                    COSMO, tracers[j], tracers[i], ell, p_of_k_a=p_of_k_a)
                assert np.allclose(corr, corr_rev)

    # Check invalid dndz
    with pytest.raises(ValueError):
        ccl.NumberCountsTracer(COSMO, has_rsd=False, dndz=z, bias=(z, b))
    with pytest.raises(ValueError):
        ccl.NumberCountsTracer(COSMO, has_rsd=False, dndz=(z, n, n),
                               bias=(z, b))
    with pytest.raises(ValueError):
        ccl.NumberCountsTracer(COSMO, has_rsd=False, dndz=(z,), bias=(z, b))
    with pytest.raises(ValueError):
        ccl.NumberCountsTracer(COSMO, has_rsd=False, dndz=(1, 2), bias=(z, b))
    with pytest.raises(ValueError):
        ccl.WeakLensingTracer(COSMO, dndz=z)
    with pytest.raises(ValueError):
        ccl.WeakLensingTracer(COSMO, dndz=(z, n, n))
    with pytest.raises(ValueError):
        ccl.WeakLensingTracer(COSMO, dndz=(z,))
    with pytest.raises(ValueError):
        ccl.WeakLensingTracer(COSMO, dndz=(1, 2))


@pytest.mark.parametrize('ells', [[3, 2, 1], [1, 3, 2], [2, 3, 1]])
def test_cells_raise_ell_reversed(ells):
    with pytest.raises(ValueError):
        ccl.angular_cl(COSMO, LENS, LENS, ells)


def test_cells_raise_integ_method():
    ells = [10, 11]
    with pytest.raises(ValueError):
        ccl.angular_cl(COSMO, LENS, LENS, ells,
                       limber_integration_method='guad')


def test_cells_raise_weird_pk():
    ells = [10, 11]
    with pytest.raises(ValueError):
        ccl.angular_cl(COSMO, LENS, LENS, ells, p_of_k_a=lambda k, a: 10)


def test_cells_mg():
    # Check that if we feed the non-linear matter power spectrum from a MG
    # cosmology into a Calculator and get Cells using MG tracers, we get the
    # same results.

    # set up a MG cosmology
    cosmo_MG = ccl.CosmologyVanillaLCDM(mg_parametrization=MuSigmaMG(
                                        mu_0=0.5, sigma_0=0.5),
                                        transfer_function="bbks",
                                        matter_power_spectrum="linear")
    cosmo_MG.compute_nonlin_power()
    pk2d = cosmo_MG.get_nonlin_power()

    # copy it into a calculator
    a, lk, pk = pk2d.get_spline_arrays()
    pk_nonlin = {"a": a, "k": np.exp(lk), "delta_matter:delta_matter": pk}
    cosmo_calc = ccl.CosmologyCalculator(
        Omega_c=0.25, Omega_b=0.05, h=0.67, n_s=0.96, sigma8=0.81,
        mg_parametrization=MuSigmaMG(mu_0=0.5, sigma_0=0.5),
        pk_nonlin=pk_nonlin)

    # get the Cells
    ell = np.geomspace(2, 2000, 128)
    tr_MG = ccl.CMBLensingTracer(cosmo_MG, z_source=1100.)
    tr_calc = ccl.CMBLensingTracer(cosmo_calc, z_source=1100.)

    cl0 = ccl.angular_cl(cosmo_MG, tr_MG, tr_MG, ell)
    cosmo_calc.compute_growth()
    cl1 = ccl.angular_cl(cosmo_calc, tr_calc, tr_calc, ell)
    assert np.all(np.fabs(1 - cl1 / cl0) < 1E-10)


ccl.gsl_params.reload()  # reset to the default parameters



import numpy as np

from ..pyutils import _get_spline1d_arrays, _get_spline2d_arrays
from .caching import _to_hashable, hash_


class Table:
    """Build nice tables.

    Comments describing the capabilities of each method are included below.
    """

    def __init__(self, *, n_y=6, n_x=6, decimals=2,
                 legend="", newline="\n\t", missing="...",
                 data_x=None, data_y=None, data_z=None, meta=[]):
        self.data_x = data_x
        self.data_y = data_y
        self.data_z = data_z
        self.meta = meta
        self.n_y = n_y
        self.n_x = n_x
        self.newline = newline
        self.missing = missing
        self.legend = legend
        self.entryl = 7 + decimals
        self.div = max(self.entryl, len(self.legend))
        self.form = f"{self.entryl}.{decimals}e"
        self.idx = np.arange(self.n_y//2).tolist() + \
            np.arange(-self.n_y//2, 0).tolist()

    def divider(self, new_line: bool):
        # Horizontal line in Table.
        # +=============+=======================+
        divider = f"+{'=' * (self.div+2)}+"
        divider += f"{'='*(1+(self.n_x*(self.entryl+1))+len(self.missing)+1)}+"
        divider += new_line * f"{self.newline}"
        return divider

    def wrap(self, expr, a, b=None):
        # Wrap the expression with `a` at the beginning and `b` at the end.
        return f"{a}{expr}{b if b is not None else a}"

    def print_left(self, expr):
        # Print the left part of a row of the Table (see `fullrow`).
        return self.wrap(self.wrap(f"{expr:{self.div}}", " "), "|")

    def print_elements(self, arr):
        # Print the elements of `arr`.
        return " ".join([f"{i:{self.form}}" for i in arr])

    def print_right(self, arr, num):
        # Print the right part of a row of the Table (see `fullrow`).
        s = self.wrap(self.print_elements(arr[:num//2]), " ")
        s += f"{self.missing}"
        s += self.wrap(self.wrap(
            self.print_elements(arr[-num//2:]),
            " "), "", f"|{self.newline}")
        return s

    def fullrow(self, s1, s2, num):
        # Print a full row of the Table.
        # |   1.00e-02  | 1.71e-01 ... 3.31e-05 |
        s = self.print_left(f"{s1:{'' if isinstance(s1, str) else self.form}}")
        s += self.print_right(s2, num)
        return s

    def missing_row(self):
        # Print row with skipped values.
        # |  ...   |            ...             |
        s = self.wrap(self.wrap(f"{self.missing:^{self.div}}", " "), "|")
        length = len(self.divider(new_line=False)) - len(s) - 3
        s += self.wrap(self.wrap(
            f"{self.missing:^{length}}",
            " "), "", f"|{self.newline}")
        return s

    def metadata(self):
        # If an object carries metadata that need to be included in `__repr__`
        # pass them here in a list. Each element will start a new row.
        s = ""
        for m in self.meta:
            s += self.wrap(self.wrap(
                f"{m:<{len(self.divider(new_line=False)) - 4}}",
                " "), "|") + self.newline
        return s

    def build(self):
        # Build the table.
        s = self.divider(new_line=True)
        s += self.fullrow(f"{self.legend}", self.data_x, self.n_x)
        s += self.divider(new_line=True)
        s += "".join([self.fullrow(self.data_y[i], self.data_z[i], self.n_x)
                      for i in self.idx[:self.n_y//2]])
        s += self.missing_row()
        s += "".join([self.fullrow(self.data_y[i], self.data_z[i], self.n_x)
                      for i in self.idx[-self.n_y//2:]])
        s += self.divider(new_line=bool(self.meta))
        s += self.metadata()
        s += self.divider(new_line=False) if self.meta else ""
        return s


def build_string_Cosmology(self):
    """Build the ``Cosmology`` representation.

    Cosmology equivalence is tested via its representation. Therefore,
    there is limiting behavior where ``'=='`` will return ``False``
    even though the compared cosmologies return the same theoretical
    predictions. This happens whenever:

        * Exactly one Cosmology is an instance of ``CosmologyCalculator``.
        * Cosmologies defined with different parameter sets, where one can
          be computed from the other (e.g. ``sigma8`` and ``A_s``).
        * Instances of ``CosmologyCalculator`` which do not contain exactly
          the same linear & non-linear power spectrum entries.

    Example output::

        <pyccl.cosmology.Cosmology>
            Omega_b = 0.05
            Omega_c = 0.25
            h       = 0.67
            n_s     = 0.96
            sigma8  = 0.81
            extra_parameters =
                test = {'param': 18.4}
            HASH_ACCURACY_PARAMS = 0x1959cbc9
            HASH_PK = 0xbca03ab0
    """
    newline = "\n\t"
    cls = self.__class__

    def test_eq(key, val, default):
        # Neutrino masses can be a list, so use `np.all` for comparison.
        # `np.all` is expensive, so only use that with `m_nu`.
        if key not in ["m_nu", "z_mg", "df_mg"]:
            return val == default
        return np.all(val == default)

    def printdict(dic):
        # Print the non-default parameters listed in a parameter dictionary.
        base = cls.__base__ if cls.__qualname__ != "Cosmology" else cls
        params = base.__signature__.parameters
        defaults = {param: value.default for param, value in params.items()}
        dic = {key: val for key, val in dic.items()
               if not test_eq(key, val, defaults.get(key))}
        dic.pop("extra_parameters", None)
        if not dic:
            return ""
        length = max(len(key) for key, val in dic.items())
        tup = _to_hashable(dic)
        s = ""
        for param, value in tup:
            s += f"{newline}{param:{length}} = {value}"
        return s

    def printextras(dic):
        # Print any extra parameters.
        if dic["extra_parameters"] is None:
            return ""
        tup = _to_hashable(dic["extra_parameters"])

        s = f"{newline}extra_parameters ="
        for key, value in tup:
            s += f"{newline}\t{key} = {dict(value)}"
        return s

    def metadata():
        # Print hashes for the accuracy parameters and the stored Pk2D's.
        H = hex(hash_(self._accuracy_params))
        s = f"{newline}HASH_ACCURACY_PARAMS = {H}"
        if self.__class__.__qualname__ == "CosmologyCalculator":
            # only need the pk's if we compare CosmologyCalculator objects
            H = 0
            if self.has_linear_power:
                H += sum([hash_(pk) for pk in self._pk_lin.values()])
            if self.has_nonlin_power:
                H += sum([hash_(pk) for pk in self._pk_nl.values()])
            H = hex(H)
            s += f"{newline}HASH_PK = {H}"
        return s

    s = "<pyccl.cosmology.Cosmology>"
    s += printdict(self._params_init_kwargs)
    s += printdict(self._config_init_kwargs)
    s += printextras(self._params_init_kwargs)
    s += metadata()
    return s


def build_string_Pk2D(self, na=6, nk=6, decimals=2):
    """Build the ``Pk2D`` representation.

    Example output ::

        <pyccl.Pk2D>
            +===============+=============================================+
            | a \\ log10(k) | -4.30e+00 -4.16e+00 ...  9.29e-01  1.02e+00 |
            +===============+=============================================+
            |   1.00e-02    |  1.71e-01  2.36e-01 ...  5.82e-05  3.31e-05 |
            |   1.26e-02    |  2.78e-01  3.82e-01 ...  9.15e-05  5.21e-05 |
            |      ...      |                     ...                     |
            |   9.77e-01    |  1.14e+03  1.57e+03 ...  3.31e-01  1.88e-01 |
            |   1.00e+00    |  1.17e+03  1.60e+03 ...  3.39e-01  1.93e-01 |
            +===============+=============================================+
            | is_log = True , extrap_orders = (1, 2)                      |
            | HASH_ARRS = 0x1d3524ad                                      |
            +===============+=============================================+
    """
    if not self.has_psp:
        return "pyccl.Pk2D(empty)"

    # get what's needed from the Pk2D object
    a, lk, pk = self.get_spline_arrays()
    lk /= np.log(10)  # easier to read in log10
    islog = str(bool(self.psp.is_log))
    extrap = (self.psp.extrap_order_lok, self.psp.extrap_order_hik)
    H = hex(sum([hash_(obj) for obj in [a, lk, pk]]))

    newline = "\n\t"  # what to do when starting a new line
    legend = "a \\ log10(k)"  # table legend
    meta = [f"is_log = {islog:5.5s}, extrap_orders = {extrap}"]
    meta += [f"HASH_ARRS = {H:34}"]

    T = Table(n_y=na, n_x=nk, decimals=decimals, legend=legend,
              newline=newline, data_x=lk, data_y=a, data_z=pk, meta=meta)

    s = build_string_simple(self) + f"{newline}"
    s += T.build()
    return s


def build_string_simple(self):
    """Simple representation.

    Example output ::

        <pyccl.emulator.Emulator>
    """
    return f"<{self.__module__}.{self.__class__.__qualname__}>"


def build_string_from_attrs(self):
    """Build a representation for an object from a list of attribute names
    given in the hook ``__repr_attrs__``.

    Example output::

        <pyccl.halos.halo_model.HMCalculator>
            mass_function = MassFuncTinker08,  HASH = 0xd3b29dd3
            halo_bias = HaloBiasTinker10,  HASH = 0x9da644b5
            mass_def = pyccl.halos.MassDef(Delta=500, rho_type=critical)
    """
    params = {param: getattr(self, param) for param in self.__repr_attrs__}
    defaults = {param: value.default
                for param, value in self.__signature__.parameters.items()
                if param != "self"}

    s = build_string_simple(self)
    newline = "\n\t"
    for param, value in params.items():
        if param in defaults and value == defaults[param]:
            # skip printing when value is the default
            continue
        s += f"{newline}{param} = "
        if "\n" in repr(value):
            # if too long, print the type and its hash
            name = value.__class__.__qualname__
            H = hex(hash_(value))
            s += f"{name},  HASH = {H}"
        else:
            s += f"{value}"
    return s


def build_string_Tracer(self):
    """Buld a representation for a Tracer.

    .. note:: Tracer insertion order is important.

    Example output ::

        <pyccl.tracers.Tracer>
            num       kernel             transfer       prefac  bessel
             0  0x82ad882c232406bb  0xa0657c0f1c98fd77    0       2
             1  0x7ab385bb323530da         None           0       0
    """
    def get_tracer_info(tr):
        # Return a string with info for the C-level tracer.

        kernel = []
        if tr.kernel is not None:
            kernel.append(_get_spline1d_arrays(tr.kernel.spline))
        kernel = hex(hash_(kernel)) if kernel else 'None'

        transfer = []
        if tr.transfer is not None:
            attrs = ["fa", "fk"]
            for attr in attrs:
                spline = getattr(tr.transfer, attr, None)
                if spline is not None:
                    transfer.append(_get_spline1d_arrays(spline))
            spline = getattr(tr.transfer, "fka", None)
            if spline is not None:
                transfer.append(_get_spline2d_arrays(spline))
            transfer.append(tr.transfer.is_log)
            transfer.append((tr.transfer.extrap_order_lok,
                             tr.transfer.extrap_order_hik))
        transfer = hex(hash_(transfer)) if transfer else 'None'

        prefac = tr.der_angles
        bessel = tr.der_bessel
        return kernel, transfer, prefac, bessel

    def print_row(newline, num, kernel, transfer, prefac, bessel):
        s = f"{num:^3}{kernel:^20}{transfer:^20}{prefac:^8}{bessel:^8}"
        return f"{newline}{s}"

    tracers = self._trc
    if not tracers:
        return "pyccl.Tracer(empty=True)"

    newline = "\n\t"
    s = build_string_simple(self)
    s += print_row(newline, "num", "kernel", "transfer", "prefac", "bessel")
    for num, tracer in enumerate(tracers):
        s += print_row(newline, num, *get_tracer_info(tracer))
    return s


def build_string_Tk3D(self, na=2, nk=4, decimals=2):
    """Build a representation for a Tk3D object.

    Example output ::

        <pyccl.Tk3D>
            +================+=============================================+
            | a \\ log10(k1) | -4.00e+00 -3.33e+00 ...  1.33e+00  2.00e+00 |
            +================+=============================================+
            |   5.00e-02     |  4.46e+07  9.62e+06 ...  2.07e+02  4.46e+01 |
            |       ...      |                     ...                     |
            |   1.00e+00     |  2.00e+09  4.30e+08 ...  9.26e+03  2.00e+03 |
            +================+=============================================+
            +================+=============================================+
            | a \\ log10(k2) | -4.00e+00 -3.33e+00 ...  1.33e+00  2.00e+00 |
            +================+=============================================+
            |   5.00e-02     |  4.46e+01  1.78e+00 ...  2.82e-10  1.12e-11 |
            |       ...      |                     ...                     |
            |   1.00e+00     |  2.00e+03  7.94e+01 ...  1.26e-08  5.01e-10 |
            +================+=============================================+
            | is_log = True , extrap_orders = (1, 1)                       |
            | HASH_ARRS = 0x780972f4                                       |
            +================+=============================================+
    """
    if not self.has_tsp:
        return "pyccl.Tk3D(empty)"

    # get what's needed from the Tk3D object
    a, lk1, lk2, tks = self.get_spline_arrays()
    lk1 /= np.log(10)  # easier to read in log10
    lk2 /= np.log(10)  # easier to read in log10
    islog = str(bool(self.tsp.is_log))
    extrap = (self.tsp.extrap_order_lok, self.tsp.extrap_order_hik)
    H = hex(sum([hash_(obj) for obj in [a, lk1, lk2, *tks]]))

    newline = "\n\t"
    meta = [f"is_log = {islog:5.5s}, extrap_orders = {extrap}"]
    meta += [f"HASH_ARRS = {H:34}"]

    # we will print 2 tables
    if not self.tsp.is_product:
        # get the start and the end of the trispectrum, diagonally in `k`
        tks = [tks[0][:, 0, :], tks[0][:, :, -1]]

    T = Table(n_y=na, n_x=nk, decimals=decimals, newline=newline,
              data_y=a, legend="a \\ log10(k1)", meta=[])

    s = build_string_simple(self) + f"{newline}"
    T.data_x, T.data_z = lk1, tks[0]
    s += T.build() + f"{newline}"
    T.legend = "a \\ log10(k2)"
    T.data_x, T.data_z = lk2, tks[1]
    T.meta = meta
    s += T.build()
    return s



import math
import pyccl as ccl

try:
    import gi
    gi.require_version('NumCosmo', '1.0')
except Exception:
    pass

from gi.repository import NumCosmo as Nc
from gi.repository import NumCosmoMath as Ncm
Ncm.cfg_init()

Omega_c = 0.25
Omega_b = 0.05
Omega_k = 0.0
h = 0.7
A_s = 2.1e-9
n_s = 0.96
Neff = 0.0
w0 = -1.0
wa = 0.0

cosmo = Nc.HICosmo.new_from_name(
    Nc.HICosmo, "NcHICosmoDECpl{'massnu-length':<0>}")
cosmo.omega_x2omega_k()
cosmo.param_set_by_name("H0", h * 100)
cosmo.param_set_by_name("Omegak", 0.0)
cosmo.param_set_by_name("w0", w0)
cosmo.param_set_by_name("w1", wa)
cosmo.param_set_by_name("Omegab", Omega_b)
cosmo.param_set_by_name("Omegac", Omega_c)
cosmo.param_set_by_name("ENnu", Neff)

# Set Omega_K in a consistent way
Omega_k = cosmo.Omega_k0()
Omega_v = cosmo.E2Omega_de(0.0)
T_CMB = cosmo.T_gamma0()
ccl_cosmo = ccl.Cosmology(
    Omega_c=Omega_c, Omega_b=Omega_b, Neff=Neff,
    h=h, n_s=n_s, Omega_k=Omega_k,
    w0=w0, wa=wa, Omega_g=0, sigma8=0.8,
    transfer_function='eisenstein_hu',
    T_CMB=T_CMB,
    matter_power_spectrum='linear')

hiprim = Nc.HIPrimPowerLaw.new()
hiprim.param_set_by_name("ln10e10ASA", math.log(1.0e10 * A_s))
hiprim.param_set_by_name("n_SA", n_s)
cosmo.add_submodel(hiprim)
dist = Nc.Distance.new(5.0)
dist.prepare(cosmo)
tf_eh = Nc.TransferFuncEH.new()
tf_eh.props.CCL_comp = True
psml = Nc.PowspecMLTransfer.new(tf_eh)
psml.require_kmin(1.0e-3)
psml.require_kmax(1.0e3)
psml.prepare(cosmo)

fact = (0.8 / psml.sigma_tophat_R(cosmo, 1.0e-7, 0.0, 8.0 / cosmo.h()))**2
hiprim.param_set_by_name("ln10e10ASA", math.log(1.0e10 * A_s * fact))
psml.require_kmin(1.0e-3)
psml.require_kmax(1.0e3)
psml.prepare(cosmo)

# now compute the mass function
psf = Ncm.PowspecFilter.new(psml, Ncm.PowspecFilterType.TOPHAT)
psf.set_best_lnr0()
psf.set_zf(2.5)
mulf = Nc.MultiplicityFunc.new_from_name("NcMultiplicityFuncTinkerMean")
mulf.set_Delta(200)
mf = Nc.HaloMassFunction.new(dist, psf, mulf)

psf.prepare(cosmo)
mf.set_area_sd(200.0)
mf.set_prec(1.0e-9)
mf.set_eval_limits(cosmo, math.log(1e14), math.log(1e16), 0.0, 2.5)
mf.prepare(cosmo)

gf = psml.peek_gf()
gf.prepare(cosmo)

cosmo.params_log_all()
print(cosmo.sigma8(psf))
print(cosmo.E(1.0/0.9 - 1))
print(psml.eval(cosmo, 1/0.9-1, 1.0))
print(gf.eval(cosmo, 1/0.9-1))

for m in [1e14, 5e14, 1e15, 5e15]:
    print(
        math.log10(m), mf.dn_dlnM(cosmo, math.log(m), 1/0.9-1) * math.log(10))

with open("../numcosmo_cluster_counts.txt", "w") as fp:
    fp.write("# counts  mmin  mmax  zmin  zmax\n")

    nc = mf.n(cosmo, math.log(1.0e14), math.log(2.0e14), 0, 2, True)
    fp.write(
        "%20.12g %20.12g %20.12g %20.12g %20.12g\n" % (nc, 1e14, 2e14, 0, 2))

    nc = mf.n(cosmo, math.log(2.0e14), math.log(1.0e15), 0, 2, True)
    fp.write(
        "%20.12g %20.12g %20.12g %20.12g %20.12g\n" % (nc, 2e14, 1e15, 0, 2))

    nc = mf.n(cosmo, math.log(1.0e15), math.log(1.0e16), 0, 2, True)
    fp.write(
        "%20.12g %20.12g %20.12g %20.12g %20.12g\n" % (nc, 1e15, 1e16, 0, 2))



__all__ = ("ConcentrationKlypin11",)

from . import Concentration


class ConcentrationKlypin11(Concentration):
    """Concentration-mass relation by `Klypin et al. 2011
    <https://arxiv.org/abs/1002.3660>`_. This parametrization is only
    valid for S.O. masses with :math:`\\Delta = \\Delta_{\\rm vir}`.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`): a mass
            definition object or a name string.
    """
    name = 'Klypin11'

    def __init__(self, *, mass_def="vir"):
        super().__init__(mass_def=mass_def)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.name != "vir"

    def _concentration(self, cosmo, M, a):
        M_pivot_inv = cosmo["h"] * 1E-12
        return 9.6 * (M * M_pivot_inv)**(-0.075)



"""Smooth background quantities

CCL defines seven species types:

* 'matter': cold dark matter and baryons
* 'dark_energy': cosmological constant or otherwise
* 'radiation': relativistic species besides massless neutrinos (i.e., only photons)
* 'curvature': curvature density
* 'neutrinos_rel': relativistic neutrinos
* 'neutrinos_massive': massive neutrinos

These strings define the `species` inputs to the functions below.
""" # noqa
__all__ = (
    "Species", "h_over_h0", "comoving_radial_distance", "scale_factor_of_chi",
    "comoving_angular_distance", "angular_diameter_distance",
    "luminosity_distance", "distance_modulus",
    "sigma_critical", "omega_x", "rho_x",
    "growth_factor", "growth_factor_unnorm", "growth_rate",)

from enum import Enum

import numpy as np

from . import lib, physical_constants
from .pyutils import (_vectorize_fn, _vectorize_fn3,
                      _vectorize_fn4, _vectorize_fn5)


class Species(Enum):
    CRITICAL = "critical"
    MATTER = "matter"
    DARK_ENERGY = "dark_energy"
    RADIATION = "radiation"
    CURVATURE = "curvature"
    NEUTRINOS_REL = "neutrinos_rel"
    NEUTRINOS_MASSIVE = "neutrinos_massive"


species_types = {
    'critical': lib.species_crit_label,
    'matter': lib.species_m_label,
    'dark_energy': lib.species_l_label,
    'radiation': lib.species_g_label,
    'curvature': lib.species_k_label,
    'neutrinos_rel': lib.species_ur_label,
    'neutrinos_massive': lib.species_nu_label,
}


def h_over_h0(cosmo, a):
    """Ratio of Hubble constant at `a` over Hubble constant today.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        a (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.

    Returns:
        (:obj:`float` or `array`): H(a)/H0.
    """
    cosmo.compute_distances()
    return _vectorize_fn(lib.h_over_h0,
                         lib.h_over_h0_vec, cosmo, a)


def comoving_radial_distance(cosmo, a):
    """Comoving radial distance.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        a (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.

    Returns:
        (:obj:`float` or `array`): Comoving radial distance; Mpc.
    """
    cosmo.compute_distances()
    return _vectorize_fn(lib.comoving_radial_distance,
                         lib.comoving_radial_distance_vec, cosmo, a)


def scale_factor_of_chi(cosmo, chi):
    """Scale factor, a, at a comoving radial distance chi.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        chi (:obj:`float` or `array`): Comoving radial distance(s); Mpc.

    Returns:
        (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.
    """
    cosmo.compute_distances()
    return _vectorize_fn(lib.scale_factor_of_chi,
                         lib.scale_factor_of_chi_vec, cosmo, chi)


def comoving_angular_distance(cosmo, a):
    """Comoving angular distance.

    .. note:: This quantity is otherwise known as the transverse
              comoving distance, and is NOT angular diameter
              distance or angular separation. The comoving angular distance
              is defined such that the comoving distance between
              two objects at a fixed scale factor separated by an angle
              :math:`\\theta` is :math:`\\theta r_{A}(a)` where
              :math:`r_{A}(a)` is the comoving angular distance.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        a (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.

    Returns:
        (:obj:`float` or `array`): Comoving angular distance; Mpc.
    """
    cosmo.compute_distances()
    return _vectorize_fn(lib.comoving_angular_distance,
                         lib.comoving_angular_distance_vec, cosmo, a)


def angular_diameter_distance(cosmo, a1, a2=None):
    """Angular diameter distance.

    The angular diameter distance in Mpc from scale factor
    `a1` to scale factor `a2`. If `a2` is not provided, it is
    assumed that the distance will be calculated between 1 and
    `a1`.

    .. note:: `a2` has to be smaller than `a1` (i.e. a source at
              `a2` is behind one at `a1`). You can compute the
              distance between a single lens at `a1` and multiple
              sources at `a2` by passing a scalar `a1`.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        a1 (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.
        a2 (:obj:`float` or `array`): Scale factor(s), normalized to 1 today,
            optional.

    Returns:
        (:obj:`float` or `array`): angular diameter distance; Mpc.
    """
    cosmo.compute_distances()
    if (a2 is not None):
        # One lens, multiple sources
        if (np.ndim(a1) == 0) and (np.ndim(a2) != 0):
            a1 = np.full(len(a2), a1)
        return _vectorize_fn5(lib.angular_diameter_distance,
                              lib.angular_diameter_distance_vec,
                              cosmo, a1, a2)
    else:
        if isinstance(a1, (int, float)):
            return _vectorize_fn5(lib.angular_diameter_distance,
                                  lib.angular_diameter_distance_vec,
                                  cosmo, 1., a1)
        else:
            return _vectorize_fn5(lib.angular_diameter_distance,
                                  lib.angular_diameter_distance_vec,
                                  cosmo, np.ones(len(a1)), a1)


def luminosity_distance(cosmo, a):
    """Luminosity distance.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        a (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.

    Returns:
        (:obj:`float` or `array`): Luminosity distance; Mpc.
    """
    cosmo.compute_distances()
    return _vectorize_fn(lib.luminosity_distance,
                         lib.luminosity_distance_vec, cosmo, a)


def distance_modulus(cosmo, a):
    """Distance Modulus, defined as

    .. math::
        \\mu = 5\\,\\log_{10}(d_L/10\\,{\\rm pc})

    where :math:`d_L` is the luminosity distance.

    .. note :: The distance modulus can be used to convert between apparent
               and absolute magnitudes via :math:`m = M + \\mu`, where
               :math:`m` is the apparent magnitude and :math:`M` is the
               absolute magnitude.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        a (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.

    Returns:
        (:obj:`float` or `array`): Distance modulus at a.
    """
    cosmo.compute_distances()
    return _vectorize_fn(lib.distance_modulus,
                         lib.distance_modulus_vec, cosmo, a)


def sigma_critical(cosmo, *, a_lens, a_source):
    """Returns the critical surface mass density.

    .. math::
         \\Sigma_{\\mathrm{crit}} = \\frac{c^2}{4\\pi G}
         \\frac{D_{\\rm{s}}}{D_{\\rm{l}}D_{\\rm{ls}}},

    where :math:`c` is the speed of light, :math:`G` is the
    gravitational constant, and :math:`D_i` is the angular diameter
    distance. The labels :math:`i = \\{s,\\,l,\\,ls\\}` denote the distances
    to the source, lens, and between source and lens, respectively.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        a_lens (:obj:`float`): lens' scale factor.
        a_source (:obj:`float` or `array`): source's scale factor.

    Returns:
        (:obj:`float` or `array`): :math:`\\Sigma_{\\mathrm{crit}}` in units
        of :math:`M_{\\odot}/{\\rm Mpc}^2`
    """
    Ds = angular_diameter_distance(cosmo, a_source, a2=None)
    Dl = angular_diameter_distance(cosmo, a_lens, a2=None)
    Dls = angular_diameter_distance(cosmo, a_lens, a_source)
    A = (
        physical_constants.CLIGHT**2
        * physical_constants.MPC_TO_METER
        / (4.0 * np.pi * physical_constants.GNEWT
           * physical_constants.SOLAR_MASS)
    )

    Sigma_crit = A * Ds / (Dl * Dls)
    return Sigma_crit


def omega_x(cosmo, a, species):
    """Density fraction of a given species at a redshift different than z=0.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        a (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.
        species (:obj:`str`): species type. Should be one of

            * 'matter': cold dark matter, massive neutrinos, and baryons
            * 'dark_energy': cosmological constant or otherwise
            * 'radiation': relativistic species besides massless neutrinos
            * 'curvature': curvature density
            * 'neutrinos_rel': relativistic neutrinos
            * 'neutrinos_massive': massive neutrinos
            * 'critical'

    Returns:
        (:obj:`float` or `array`): Density fraction of a given species at a \
                             scale factor.
    """
    # TODO: Replace docstring enum with ref to Species.
    if species not in species_types:
        raise ValueError(f"Unknown species {species}.")

    return _vectorize_fn3(lib.omega_x,
                          lib.omega_x_vec, cosmo, a, species_types[species])


def rho_x(cosmo, a, species, *, is_comoving=False):
    """Physical or comoving density as a function of scale factor.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        a (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.
        species (:obj:`str`): species type. Should be one of

            - 'matter': cold dark matter, massive neutrinos, and baryons
            - 'dark_energy': cosmological constant or otherwise
            - 'radiation': relativistic species besides massless neutrinos
            - 'curvature': curvature density
            - 'neutrinos_rel': relativistic neutrinos
            - 'neutrinos_massive': massive neutrinos
            - 'critical'

        is_comoving (:obj:`bool`): either physical (False, default) or
            comoving (True)

    Returns:
        (:obj:`float` or `array`): Physical density of a given species
        at a scale factor, in units of :math:`M_\\odot / {\\rm Mpc}^3`.
    """
    # TODO: Replace docstring enum with ref to Species.
    if species not in species_types:
        raise ValueError(f"Unknown species {species}.")

    return _vectorize_fn4(
        lib.rho_x, lib.rho_x_vec, cosmo, a,
        species_types[species], int(is_comoving))


def growth_factor(cosmo, a):
    """Growth factor.

    .. warning:: CCL is not able to compute the scale-dependent growth
                 factor for cosmologies with massive neutrinos.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        a (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.

    Returns:
        (:obj:`float` or `array`): Growth factor, normalized to unity today.
    """
    cosmo.compute_growth()
    return _vectorize_fn(lib.growth_factor,
                         lib.growth_factor_vec, cosmo, a)


def growth_factor_unnorm(cosmo, a):
    """Unnormalized growth factor.

    .. warning:: CCL is not able to compute the scale-dependent growth
                 factor for cosmologies with massive neutrinos.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        a (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.

    Returns:
        (:obj:`float` or `array`): Unnormalized growth factor, normalized to \
            the scale factor at early times.
    """
    cosmo.compute_growth()
    return _vectorize_fn(lib.growth_factor_unnorm,
                         lib.growth_factor_unnorm_vec, cosmo, a)


def growth_rate(cosmo, a):
    """Growth rate defined as the logarithmic derivative of the
    growth factor, :math:`f\\equiv d\\log D/d\\log a`.

    .. warning:: CCL is not able to compute the scale-dependent growth
                 rate for cosmologies with massive neutrinos.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
        a (:obj:`float` or `array`): Scale factor(s), normalized to 1 today.

    Returns:
        (:obj:`float` or `array`): Growth rate.

    """
    cosmo.compute_growth()
    return _vectorize_fn(lib.growth_rate,
                         lib.growth_rate_vec, cosmo, a)



__all__ = ("halomod_trispectrum_1h", "halomod_Tk3D_1h",
           "halomod_Tk3D_SSC_linear_bias", "halomod_Tk3D_SSC",)

import warnings

import numpy as np

from .. import CCLWarning, Tk3D
from . import HaloProfileNFW, Profile2pt


def halomod_trispectrum_1h(cosmo, hmc, k, a, prof, *,
                           prof2=None, prof3=None, prof4=None,
                           prof12_2pt=None, prof34_2pt=None):
    """ Computes the halo model 1-halo trispectrum for four different
    quantities defined by their respective halo profiles. The 1-halo
    trispectrum for four profiles :math:`u_{1,2}`, :math:`v_{1,2}` is
    calculated as:

    .. math::
        T_{u_1,u_2;v_1,v_2}(k_u,k_v,a) =
        I^0_{2,2}(k_u,k_v,a|u_{1,2},v_{1,2})

    where :math:`I^0_{2,2}` is defined in the documentation
    of :meth:`~pyccl.halos.halo_model.HMCalculator.I_0_22`.

    .. note:: This approximation assumes that the 4-point
              profile cumulant is the same as the product of two
              2-point cumulants. We may relax this assumption in
              future versions of CCL.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        hmc (:class:`HMCalculator`): a halo model calculator.
        k (:obj:`float` or `array`): comoving wavenumber in
            :math:`{\\rm Mpc}^{-1}`.
        a (:obj:`float` or `array`): scale factor.
        prof (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile (corresponding to :math:`u_1` above).
        prof2 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile (corresponding to :math:`u_2` above). If ``None``,
            ``prof`` will be used as ``prof2``.
        prof12_2pt (:class:`~pyccl.halos.profiles_2pt.Profile2pt`):
            a profile covariance object returning the the two-point
            moment of ``prof`` and ``prof2``. If ``None``, the default
            second moment will be used, corresponding to the
            products of the means of both profiles.
        prof3 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile (corresponding to :math:`v_1` above. If ``None``,
            ``prof`` will be used as ``prof3``.
        prof4 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile (corresponding to :math:`v_2` above. If ``None``,
            ``prof2`` will be used as ``prof4``.
        prof34_2pt (:class:`~pyccl.halos.profiles_2pt.Profile2pt`):
            same as ``prof12_2pt`` for ``prof3`` and ``prof4``.

    Returns:
        (:obj:`float` or `array`): 1-halo trispectrum evaluated at each
        combination of ``k`` and ``a``. The shape of the output will
        be ``(N_a, N_k, N_k)`` where ``N_k`` and ``N_a`` are the sizes of
        ``k`` and ``a`` respectively. The ordering is such that
        ``output[ia, ik2, ik1] = T(k[ik1], k[ik2], a[ia])``
        If ``k`` or ``a`` are scalars, the corresponding dimension will
        be squeezed out on output.
    """
    a_use = np.atleast_1d(a).astype(float)
    k_use = np.atleast_1d(k).astype(float)

    # define all the profiles
    prof, prof2, prof3, prof4, prof12_2pt, prof34_2pt = \
        _allocate_profiles(prof, prof2, prof3, prof4, prof12_2pt, prof34_2pt)

    na = len(a_use)
    nk = len(k_use)
    out = np.zeros([na, nk, nk])
    for ia, aa in enumerate(a_use):
        # normalizations
        norm1 = prof.get_normalization(cosmo, aa, hmc=hmc)

        if prof2 == prof:
            norm2 = norm1
        else:
            norm2 = prof2.get_normalization(cosmo, aa, hmc=hmc)

        if prof3 == prof:
            norm3 = norm1
        else:
            norm3 = prof3.get_normalization(cosmo, aa, hmc=hmc)

        if prof4 == prof2:
            norm4 = norm2
        else:
            norm4 = prof4.get_normalization(cosmo, aa, hmc=hmc)

        # trispectrum
        tk_1h = hmc.I_0_22(cosmo, k_use, aa,
                           prof=prof, prof2=prof2,
                           prof4=prof4, prof3=prof3,
                           prof12_2pt=prof12_2pt,
                           prof34_2pt=prof34_2pt)

        out[ia] = tk_1h / (norm1 * norm2 * norm3 * norm4)  # assign

    if np.ndim(a) == 0:
        out = np.squeeze(out, axis=0)
    if np.ndim(k) == 0:
        out = np.squeeze(out, axis=-1)
        out = np.squeeze(out, axis=-1)
    return out


def halomod_Tk3D_1h(cosmo, hmc, prof, *,
                    prof2=None, prof3=None, prof4=None,
                    prof12_2pt=None, prof34_2pt=None,
                    lk_arr=None, a_arr=None,
                    extrap_order_lok=1, extrap_order_hik=1,
                    use_log=False):
    """ Returns a :class:`~pyccl.tk3d.Tk3D` object containing
    the 1-halo trispectrum for four quantities defined by
    their respective halo profiles. See :meth:`halomod_trispectrum_1h`
    for more details about the actual calculation.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        hmc (:class:`HMCalculator`): a halo model calculator.
        prof (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile (corresponding to :math:`u_1` above.
        prof2 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile. If ``None``, ``prof`` will be used as ``prof2``.
        prof3 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile. If ``None``, ``prof`` will be used as ``prof3``.
        prof4 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile. If ``None``, ``prof2`` will be used as ``prof4``.
        prof12_2pt (:class:`~pyccl.halos.profiles_2pt.Profile2pt`):
            a profile covariance object returning the the two-point
            moment of ``prof`` and ``prof2``. If ``None``, the default
            second moment will be used, corresponding to the
            products of the means of both profiles.
        prof34_2pt (:class:`~pyccl.halos.profiles_2pt.Profile2pt`):
            same as ``prof12_2pt`` for ``prof3`` and ``prof4``.
        a_arr (array): an array holding values of the scale factor
            at which the trispectrum should be calculated for
            interpolation. If ``None``, the internal values used
            by ``cosmo`` will be used.
        lk_arr (array): an array holding values of the natural
            logarithm of the wavenumber (in units of
            :math:`{\\rm Mpc}^{-1}`) at which the trispectrum should
            be calculated for interpolation. If ``None``, the internal
            values used by ``cosmo`` will be used.
        extrap_order_lok (:obj:`int`): extrapolation order to be used on
            k-values below the minimum of the splines. See
            :class:`~pyccl.tk3d.Tk3D`.
        extrap_order_hik (:obj:`int`): extrapolation order to be used on
            k-values above the maximum of the splines. See
            :class:`~pyccl.tk3d.Tk3D`.
        use_log (:obj:`bool`): if ``True``, the trispectrum will be
            interpolated in log-space (unless negative or
            zero values are found).

    Returns:
        :class:`~pyccl.tk3d.Tk3D`: 1-halo trispectrum.
    """
    if lk_arr is None:
        lk_arr = cosmo.get_pk_spline_lk()
    if a_arr is None:
        a_arr = cosmo.get_pk_spline_a()

    tkk = halomod_trispectrum_1h(cosmo, hmc, np.exp(lk_arr), a_arr,
                                 prof, prof2=prof2,
                                 prof12_2pt=prof12_2pt,
                                 prof3=prof3, prof4=prof4,
                                 prof34_2pt=prof34_2pt)

    tkk, use_log = _logged_output(tkk, log=use_log)

    return Tk3D(a_arr=a_arr, lk_arr=lk_arr, tkk_arr=tkk,
                extrap_order_lok=extrap_order_lok,
                extrap_order_hik=extrap_order_hik, is_logt=use_log)


def halomod_Tk3D_SSC_linear_bias(cosmo, hmc, *, prof,
                                 bias1=1, bias2=1, bias3=1, bias4=1,
                                 is_number_counts1=False,
                                 is_number_counts2=False,
                                 is_number_counts3=False,
                                 is_number_counts4=False,
                                 p_of_k_a=None, lk_arr=None,
                                 a_arr=None, extrap_order_lok=1,
                                 extrap_order_hik=1, use_log=False,
                                 extrap_pk=False):
    """ Returns a :class:`~pyccl.tk3d.Tk3D` object containing
    the super-sample covariance trispectrum, given by the tensor
    product of the power spectrum responses associated with the
    two pairs of quantities being correlated. Each response is
    calculated as:

    .. math::
        \\frac{\\partial P_{u,v}(k)}{\\partial\\delta_L} = b_u b_v \\left(
        \\left(\\frac{68}{21}-\\frac{d\\log k^3P_L(k)}{d\\log k}\\right)
        P_L(k)+I^1_2(k|u,v)\\right) - (b_{u} + b_{v}) P_{u,v}(k)

    where the :math:`I^1_2` is defined in the documentation
    :meth:`~pyccl.halos.halo_model.HMCalculator.I_1_2` and :math:`b_{u}`
    and :math:`b_{v}` are the linear halo biases for quantities :math:`u`
    and :math:`v`, respectively. The second term is only included if the
    corresponding profiles do not represent number counts.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        hmc (:class:`HMCalculator`): a halo model calculator.
        prof (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): a halo
            profile representing the matter overdensity.
        bias1 (:obj:`float` or `array`): linear galaxy bias for quantity 1.
            If an array, it has to have the shape of ``a_arr``.
        bias2 (:obj:`float` or `array`): linear galaxy bias for quantity 2.
        bias3 (:obj:`float` or `array`): linear galaxy bias for quantity 3.
        bias4 (:obj:`float` or `array`): linear galaxy bias for quantity 4.
        is_number_counts1 (:obj:`bool`): If ``True``, quantity 1 will be considered
            number counts and the clustering counter terms computed.
        is_number_counts2 (:obj:`bool`): as ``is_number_counts1`` but for quantity 2.
        is_number_counts3 (:obj:`bool`): as ``is_number_counts1`` but for quantity 3.
        is_number_counts4 (:obj:`bool`): as ``is_number_counts1`` but for quantity 4.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`): a `Pk2D` object to
            be used as the linear matter power spectrum. If ``None``,
            the power spectrum stored within ``cosmo`` will be used.
        a_arr (array): an array holding values of the scale factor
            at which the trispectrum should be calculated for
            interpolation. If ``None``, the internal values used
            by ``cosmo`` will be used.
        lk_arr (array): an array holding values of the natural
            logarithm of the wavenumber (in units of
            :math:`{\\rm Mpc}^{-1}`) at which the trispectrum should be
            calculated for interpolation. If ``None``, the internal values
            used by ``cosmo`` will be used.
        extrap_order_lok (:obj:`int`): extrapolation order to be used on
            k-values below the minimum of the splines. See
            :class:`~pyccl.tk3d.Tk3D`.
        extrap_order_hik (:obj:`int`): extrapolation order to be used on
            k-values above the maximum of the splines. See
            :class:`~pyccl.tk3d.Tk3D`.
        use_log (:obj:`bool`): if ``True``, the trispectrum will be
            interpolated in log-space (unless negative or
            zero values are found).
        extrap_pk (:obj:`bool`):
            Whether to extrapolate ``p_of_k_a`` in case ``a`` is out of its
            support. If ``False``, and the queried values are out of bounds,
            an error is raised. The default is ``False``.

    Returns:
        :class:`~pyccl.tk3d.Tk3D`: SSC effective trispectrum.
    """ # noqa
    if lk_arr is None:
        lk_arr = cosmo.get_pk_spline_lk()
    if a_arr is None:
        a_arr = cosmo.get_pk_spline_a()

    if not isinstance(prof, HaloProfileNFW):
        raise TypeError("prof should be HaloProfileNFW.")

    # Make sure biases are of the form number of a x number of k
    ones = np.ones_like(a_arr)
    bias1 *= ones
    bias2 *= ones
    bias3 *= ones
    bias4 *= ones

    k_use = np.exp(lk_arr)
    prof_2pt = Profile2pt()

    pk2d = cosmo.parse_pk(p_of_k_a)
    extrap = cosmo if extrap_pk else None  # extrapolation rule for pk2d

    na = len(a_arr)
    nk = len(k_use)
    dpk12, dpk34 = [np.zeros([na, nk]) for _ in range(2)]
    for ia, aa in enumerate(a_arr):
        norm = prof.get_normalization(cosmo, aa, hmc=hmc)**2
        i12 = hmc.I_1_2(cosmo, k_use, aa, prof, prof2=prof, prof_2pt=prof_2pt)

        pk = pk2d(k_use, aa, cosmo=extrap)
        dpk = pk2d(k_use, aa, derivative=True, cosmo=extrap)

        # ~ (47/21 - 1/3 dlogPk/dlogk) * Pk + I12
        dpk12[ia] = (47/21 - dpk/3)*pk + i12 / norm
        dpk34[ia] = dpk12[ia].copy()

        # Counter terms for clustering (i.e. - (bA + bB) * PAB)
        if any([is_number_counts1, is_number_counts2,
                is_number_counts3, is_number_counts4]):
            b1 = b2 = b3 = b4 = 0
            i02 = hmc.I_0_2(cosmo, k_use, aa, prof,
                            prof2=prof, prof_2pt=prof_2pt)

            P_12 = P_34 = pk + i02 / norm

            if is_number_counts1:
                b1 = bias1[ia]
            if is_number_counts2:
                b2 = bias2[ia]
            if is_number_counts3:
                b3 = bias3[ia]
            if is_number_counts4:
                b4 = bias4[ia]

            dpk12[ia] -= (b1 + b2) * P_12
            dpk34[ia] -= (b3 + b4) * P_34

        dpk12[ia] *= bias1[ia] * bias2[ia]
        dpk34[ia] *= bias3[ia] * bias4[ia]

    dpk12, dpk34, use_log = _logged_output(dpk12, dpk34, log=use_log)

    return Tk3D(a_arr=a_arr, lk_arr=lk_arr,
                pk1_arr=dpk12, pk2_arr=dpk34,
                extrap_order_lok=extrap_order_lok,
                extrap_order_hik=extrap_order_hik, is_logt=use_log)


def halomod_Tk3D_SSC(
        cosmo, hmc, prof, *, prof2=None, prof3=None, prof4=None,
        prof12_2pt=None, prof34_2pt=None,
        p_of_k_a=None, lk_arr=None, a_arr=None,
        extrap_order_lok=1, extrap_order_hik=1, use_log=False,
        extrap_pk=False):
    """ Returns a :class:`~pyccl.tk3d.Tk3D` object containing
    the super-sample covariance trispectrum, given by the tensor
    product of the power spectrum responses associated with the
    two pairs of quantities being correlated. Each response is
    calculated as:

    .. math::
        \\frac{\\partial P_{u,v}(k)}{\\partial\\delta_L} =
        \\left(\\frac{68}{21}-\\frac{d\\log k^3P_L(k)}{d\\log k}\\right)
        P_L(k)I^1_1(k,|u)I^1_1(k,|v)+I^1_2(k|u,v) - (b_{u} + b_{v})
        P_{u,v}(k)

    where the :math:`I^a_b` are defined in the documentation
    of :meth:`~pyccl.halos.halo_model.HMCalculator.I_1_1` and
    :meth:`~pyccl.halos.halo_model.HMCalculator.I_1_2` and
    :math:`b_{u}` and :math:`b_{v}` are the linear halo biases for
    quantities :math:`u` and :math:`v`, respectively (zero if the
    profiles are not number counts).

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        hmc (:class:`~pyccl.halos.halo_model.HMCalculator`):
            a halo model calculator.
        prof (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile.
        prof2 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile (corresponding to :math:`u_2` above). If ``None``,
            ``prof`` will be used as ``prof2``.
        prof3 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile (corresponding to :math:`v_1` above). If ``None``,
            ``prof`` will be used as ``prof3``.
        prof4 (:class:`~pyccl.halos.profiles.profile_base.HaloProfile`): halo
            profile (corresponding to :math:`v_2` above). If ``None``,
            ``prof2`` will be used as ``prof4``.
        prof12_2pt (:class:`~pyccl.halos.profiles_2pt.Profile2pt`):
            a profile covariance object returning the the two-point
            moment of ``prof`` and ``prof2``. If ``None``, the default
            second moment will be used, corresponding to the
            products of the means of both profiles.
        prof34_2pt (:class:`~pyccl.halos.profiles_2pt.Profile2pt`):
            same as ``prof12_2pt`` for ``prof3`` and ``prof4``. If ``None``,
            ``prof12_2pt`` will be used.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D`): a `Pk2D` object to
            be used as the linear matter power spectrum. If ``None``,
            the power spectrum stored within ``cosmo`` will be used.
        a_arr (array): an array holding values of the scale factor
            at which the trispectrum should be calculated for
            interpolation. If ``None``, the internal values used
            by ``cosmo`` will be used.
        lk_arr (array): an array holding values of the natural
            logarithm of the wavenumber (in units of
            :math:`{\\rm Mpc}^{-1}`) at which the trispectrum should
            be calculated for interpolation. If ``None``, the internal
            values used by ``cosmo`` will be used.
        extrap_order_lok (:obj:`int`): extrapolation order to be used on
            k-values below the minimum of the splines. See
            :class:`~pyccl.tk3d.Tk3D`.
        extrap_order_hik (:obj:`int`): extrapolation order to be used on
            k-values above the maximum of the splines. See
            :class:`~pyccl.tk3d.Tk3D`.
        use_log (:obj:`bool`): if ``True``, the trispectrum will be
            interpolated in log-space (unless negative or
            zero values are found).
        extrap_pk (:obj:`bool`):
            Whether to extrapolate ``p_of_k_a`` in case ``a`` is out of its
            support. If ``False``, and the queried values are out of bounds,
            an error is raised. The default is ``False``.

    Returns:
        :class:`~pyccl.tk3d.Tk3D`: SSC effective trispectrum.
    """
    if lk_arr is None:
        lk_arr = cosmo.get_pk_spline_lk()
    if a_arr is None:
        a_arr = cosmo.get_pk_spline_a()

    # define all the profiles
    prof, prof2, prof3, prof4, prof12_2pt, prof34_2pt = \
        _allocate_profiles(prof, prof2, prof3, prof4, prof12_2pt, prof34_2pt)

    k_use = np.exp(lk_arr)
    pk2d = cosmo.parse_pk(p_of_k_a)
    extrap = cosmo if extrap_pk else None  # extrapolation rule for pk2d

    dpk12, dpk34 = [np.zeros((len(a_arr), len(k_use))) for _ in range(2)]
    for ia, aa in enumerate(a_arr):
        # normalizations & I11 integral
        norm1 = prof.get_normalization(cosmo, aa, hmc=hmc)
        i11_1 = hmc.I_1_1(cosmo, k_use, aa, prof)

        if prof2 == prof:
            norm2 = norm1
            i11_2 = i11_1
        else:
            norm2 = prof2.get_normalization(cosmo, aa, hmc=hmc)
            i11_2 = hmc.I_1_1(cosmo, k_use, aa, prof2)

        if prof3 == prof:
            norm3 = norm1
            i11_3 = i11_1
        else:
            norm3 = prof3.get_normalization(cosmo, aa, hmc=hmc)
            i11_3 = hmc.I_1_1(cosmo, k_use, aa, prof3)

        if prof4 == prof2:
            norm4 = norm2
            i11_4 = i11_2
        else:
            norm4 = prof4.get_normalization(cosmo, aa, hmc=hmc)
            i11_4 = hmc.I_1_1(cosmo, k_use, aa, prof4)

        # I12 integral
        i12_12 = hmc.I_1_2(cosmo, k_use, aa, prof,
                           prof2=prof2, prof_2pt=prof12_2pt)
        if (prof, prof2, prof12_2pt) == (prof3, prof4, prof34_2pt):
            i12_34 = i12_12
        else:
            i12_34 = hmc.I_1_2(cosmo, k_use, aa, prof3,
                               prof2=prof4, prof_2pt=prof34_2pt)

        # power spectrum
        pk = pk2d(k_use, aa, cosmo=extrap)
        dpk = pk2d(k_use, aa, derivative=True, cosmo=extrap)

        # (47/21 - 1/3 dlogPk/dlogk) * I11 * I11 * Pk + I12
        dpk12[ia] = ((47/21 - dpk/3)*i11_1*i11_2*pk + i12_12) / (norm1 * norm2)
        dpk34[ia] = ((47/21 - dpk/3)*i11_3*i11_4*pk + i12_34) / (norm3 * norm4)

        # Counter terms for clustering (i.e. - (bA + bB) * PAB)
        def _get_counterterm(pA, pB, p2pt, nA, nB, i11_A, i11_B):
            """Helper to compute counter-terms."""
            # p : profiles | p2pt : 2-point | n : norms | i11 : I_1_1 integral
            bA = i11_A / nA if pA.is_number_counts else np.zeros_like(k_use)
            bB = i11_B / nB if pB.is_number_counts else np.zeros_like(k_use)
            i02 = hmc.I_0_2(cosmo, k_use, aa, pA, prof2=pB, prof_2pt=p2pt)
            P = (pk * i11_A * i11_B + i02) / (nA * nB)
            return (bA + bB) * P

        if prof.is_number_counts or prof2.is_number_counts:
            dpk12[ia] -= _get_counterterm(prof, prof2, prof12_2pt,
                                          norm1, norm2, i11_1, i11_2)

        if prof3.is_number_counts or prof4.is_number_counts:
            if (prof, prof2, prof12_2pt) == (prof3, prof4, prof34_2pt):
                dpk34[ia] = dpk12[ia]
            else:
                dpk34[ia] -= _get_counterterm(prof3, prof4, prof34_2pt,
                                              norm3, norm4, i11_3, i11_4)

    dpk12, dpk34, use_log = _logged_output(dpk12, dpk34, log=use_log)

    return Tk3D(a_arr=a_arr, lk_arr=lk_arr,
                pk1_arr=dpk12, pk2_arr=dpk34,
                extrap_order_lok=extrap_order_lok,
                extrap_order_hik=extrap_order_hik, is_logt=use_log)


def _allocate_profiles(prof, prof2, prof3, prof4, prof12_2pt, prof34_2pt):
    """Helper that controls how the undefined profiles are allocated."""
    if prof2 is None:
        prof2 = prof
    if prof3 is None:
        prof3 = prof
    if prof4 is None:
        prof4 = prof2
    if prof12_2pt is None:
        prof12_2pt = Profile2pt()
    if prof34_2pt is None:
        prof34_2pt = prof12_2pt

    return prof, prof2, prof3, prof4, prof12_2pt, prof34_2pt


def _logged_output(*arrs, log):
    """Helper that logs the output if needed."""
    if not log:
        return *arrs, log
    is_negative = [(arr <= 0).any() for arr in arrs]
    if any(is_negative):
        warnings.warn("Some values were non-positive. "
                      "Interpolating linearly.", CCLWarning)
        return *arrs, False
    return *[np.log(arr) for arr in arrs], log



__all__ = ("MassFuncTinker08",)

import numpy as np
from scipy.interpolate import interp1d

from . import MassFunc


class MassFuncTinker08(MassFunc):
    """Implements the mass function of `Tinker et al. 2008
    <https://arxiv.org/abs/0803.2706>`_. This parametrization accepts S.O.
    masses with :math:`200 < \\Delta < 3200`, defined with respect to the
    matter density. This can be automatically translated to S.O. masses
    defined with respect to the critical density.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
    """
    name = 'Tinker08'

    def __init__(self, *,
                 mass_def="200m",
                 mass_def_strict=True):
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.Delta == "fof"

    def _setup(self):
        delta = np.array(
            [200., 300., 400., 600., 800., 1200., 1600., 2400., 3200.])
        alpha = np.array(
            [0.186, 0.200, 0.212, 0.218, 0.248, 0.255, 0.260, 0.260, 0.260])
        beta = np.array(
            [1.47, 1.52, 1.56, 1.61, 1.87, 2.13, 2.30, 2.53, 2.66])
        gamma = np.array(
            [2.57, 2.25, 2.05, 1.87, 1.59, 1.51, 1.46, 1.44, 1.41])
        phi = np.array(
            [1.19, 1.27, 1.34, 1.45, 1.58, 1.80, 1.97, 2.24, 2.44])
        ldelta = np.log10(delta)
        self.pA0 = interp1d(ldelta, alpha)
        self.pa0 = interp1d(ldelta, beta)
        self.pb0 = interp1d(ldelta, gamma)
        self.pc = interp1d(ldelta, phi)

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        ld = np.log10(self.mass_def._get_Delta_m(cosmo, a))
        pA = self.pA0(ld) * a**0.14
        pa = self.pa0(ld) * a**0.06
        pd = 10.**(-(0.75/(ld - 1.8750612633))**1.2)
        pb = self.pb0(ld) * a**pd
        return pA * ((pb / sigM)**pa + 1) * np.exp(-self.pc(ld)/sigM**2)



import numpy as np
import matplotlib.pyplot as plt
import py_cosmo_mad as csm

#Produces all distance benchmarks
#Contact david.alonso@physics.ox.ac.uk if you have issues running this script

TCMB=2.725
PLOT_STUFF=0
WRITE_STUFF=1
FS=16

def do_all(z_arr,cpar,prefix) :
    pcs=csm.PcsPar()
    pcs.background_set(cpar['om'],cpar['ol'],cpar['ob'],cpar['w0'],cpar['wa'],cpar['hh'],TCMB)

    a_arr=1./(z_arr+1)
    chi_arr=np.array([pcs.radial_comoving_distance(a) for a in a_arr])

    if PLOT_STUFF==1 :
        plt.plot(z_arr,chi_arr); plt.xlabel('$z$',fontsize=FS); plt.ylabel('$\\chi(z)\\,[{\\rm Mpc}\\,h^{-1}]$',fontsize=FS); plt.show()

    if WRITE_STUFF==1 :
        np.savetxt(prefix+"_chi.txt",np.transpose([z_arr,chi_arr]),header="[1] z, [2] chi(z) (Mpc/h)")

z_arr=np.array([0.,1.,2.,3.,4.,5.])

cpar_model1={'om': 0.3,'ol': 0.7 ,'ob':0.05,'hh': 0.7,'s8': 0.8,'ns': 0.96,'w0': -1.0, 'wa': 0.0}
cpar_model2={'om': 0.3,'ol': 0.7 ,'ob':0.05,'hh': 0.7,'s8': 0.8,'ns': 0.96,'w0': -0.9, 'wa': 0.0}
cpar_model3={'om': 0.3,'ol': 0.7 ,'ob':0.05,'hh': 0.7,'s8': 0.8,'ns': 0.96,'w0': -0.9, 'wa': 0.1}
cpar_model4={'om': 0.3,'ol': 0.75,'ob':0.05,'hh': 0.7,'s8': 0.8,'ns': 0.96,'w0': -0.9, 'wa': 0.1}
cpar_model5={'om': 0.3,'ol': 0.65,'ob':0.05,'hh': 0.7,'s8': 0.8,'ns': 0.96,'w0': -0.9, 'wa': 0.1}

do_all(z_arr,cpar_model1,"model1")
do_all(z_arr,cpar_model2,"model2")
do_all(z_arr,cpar_model3,"model3")
do_all(z_arr,cpar_model3,"model4")
do_all(z_arr,cpar_model3,"model5")



import os
#Produces all angular power spectrum benchmarks
#Contact david.alonso@physics.ox.ac.uk if you have issues running this script

def run_task(window_type,bins_type,b1,b2) :
    do_logbin=0
    if bins_type=="lin" :
        do_logbin=0
    elif bins_type=="log" :
        do_logbin=1
    else :
        print("Wrong binning scheme "+bins_type)

    strout="#Cosmological parameters\n"
    strout+="omega_m= 0.3\n"
    strout+="omega_l= 0.7\n"
    strout+="omega_b= 0.0\n"
    strout+="w0= -1.\n"
    strout+="wa= 0.\n"
    strout+="h= 0.7\n"
    strout+="ns= 0.96\n"
    strout+="s8= 0.8\n"
    strout+="\n"
    strout+="#Radial resolution\n"
    strout+="d_chi= 5.\n"
    strout+="\n"
    strout+="#Maximum multipole\n"
    strout+="l_max= 30000\n"
    strout+="\n"
    strout+="#Behavioural flags (include number counts? include lensing shear? include CMB lensing?)\n"
    strout+="do_nc= 1\n"
    strout+="has_nc_dens= 1\n"
    strout+="has_nc_rsd= 0\n"
    strout+="has_nc_lensing= 0\n"
    strout+="do_shear= 1\n"
    strout+="do_cmblens= 1\n"
    strout+="\n"
    strout+="#Angular correlation function\n"
    strout+="do_w_theta= 1\n"
    strout+="use_logbin= %d\n"%do_logbin
    strout+="theta_min= 0\n"
    strout+="theta_max= 10.\n"
    strout+="n_bins_theta= 15\n"
    strout+="n_bins_decade= 5\n"
    strout+="\n"
    strout+="#File names (window function, bias, magnification bias, power spectrum)\n"
    strout+="window_1_fname= curves/bin%d_"%b1+window_type+".txt\n"
    strout+="window_2_fname= curves/bin%d_"%b2+window_type+".txt\n"
    strout+="bias_fname= curves/bias.txt\n"
    strout+="sbias_fname= nothing\n"
    strout+="pk_fname= BBKS\n"
    strout+="\n"
    strout+="#Output prefix\n"
    strout+="prefix_out= output/run_b%d"%b1+"b%d"%b2+window_type+"_"+bins_type

    parname="param_limberjack_b%d"%b1+"b%d"%b2+window_type+"_"+bins_type+".ini"
    f=open(parname,"w")
    f.write(strout)
    f.close()

    os.system("./limberjack "+parname)

run_task("analytic","log",1,1)
run_task("analytic","log",1,2)
run_task("analytic","log",2,2)
run_task("histo","log",1,1)
run_task("histo","log",1,2)
run_task("histo","log",2,2)



import numpy as np
import pytest
from pyccl.pyutils import _spline_integrate
from pyccl.errors import CCLError


def yf(x, pw):
    return x**pw


def yf_int(x, pw):
    return x**(pw + 1) / (pw + 1)


def test_intspline_accuracy():
    nx = 1024
    x_arr = np.linspace(-2, 2, nx)
    y_arr = np.array([yf(x_arr, p)
                      for p in [2, 4]])
    int_spline = _spline_integrate(x_arr, y_arr, -1, 1)
    int_exact = np.array([yf_int(1, p) - yf_int(-1, p)
                          for p in [2, 4]])
    assert np.all(np.fabs(int_spline / int_exact - 1) < 1E5-5)


def test_intspline_raises_shapes():
    x = np.linspace(0, 1, 10)
    y = np.zeros([4, 10])

    # x is 0- or 2-D
    with pytest.raises(ValueError):
        _spline_integrate(x[0], y, 0, 1)

    with pytest.raises(ValueError):
        _spline_integrate(y, y, 0, 1)

    # y is 0-D
    with pytest.raises(ValueError):
        _spline_integrate(x, y[0, 0], 0, 1)

    # Mismatching x and y
    with pytest.raises(ValueError):
        _spline_integrate(x[1:], y, 0, 1)

    # Limits as arrays
    with pytest.raises(TypeError):
        _spline_integrate(x, y, np.zeros(10), 1)


def test_intspline_raises_internal():
    x = np.linspace(0, 1, 10)
    y = np.zeros([4, 10])

    # Function is not splineable
    with pytest.raises(CCLError):
        _spline_integrate(0 * x, y, 0, 1)


def test_intspline_smoke():
    x = np.linspace(0, 1, 10)
    y = yf(x, 2)

    r = _spline_integrate(x, y, 0, 1)
    assert np.all(~np.isnan(r))

    r = _spline_integrate(x, np.array([y, y]), 0, 1)
    assert np.all(~np.isnan(r))



import numpy as np
import pyccl as ccl
import pytest


COSMO = ccl.CosmologyVanillaLCDM(transfer_function='bbks')
bar = ccl.BaryonsSchneider15()


@pytest.mark.parametrize('k', [
    1,
    1.0,
    [0.3, 0.5, 10],
    np.array([0.3, 0.5, 10])])
def test_bcm_smoke(k):
    a = 0.8
    fka = bar.boost_factor(COSMO, k, a)
    assert np.all(np.isfinite(fka))
    assert np.shape(fka) == np.shape(k)


def test_bcm_correct_smoke():
    k_arr = np.geomspace(1E-2, 1, 10)
    fka = bar.boost_factor(COSMO, k_arr, 0.5)
    pk_nobar = ccl.nonlin_matter_power(COSMO, k_arr, 0.5)
    pkb = bar.include_baryonic_effects(
        COSMO, COSMO.get_nonlin_power())
    pk_wbar = pkb(k_arr, 0.5)
    assert np.all(np.fabs(pk_wbar/(pk_nobar*fka)-1) < 1E-5)


def test_bcm_update_params():
    bar2 = ccl.BaryonsSchneider15(log10Mc=14.1, eta_b=0.7, k_s=40.)
    bar2.update_parameters(log10Mc=bar.log10Mc,
                           eta_b=bar.eta_b,
                           k_s=bar.k_s)
    assert bar == bar2


def test_baryons_from_name():
    bar2 = ccl.Baryons.from_name('Schneider15')
    assert bar.name == bar2.name
    assert bar2.name == 'Schneider15'


def test_baryons_in_cosmology():
    # Test that applying baryons during cosmology creation works.
    # 1. Outside of cosmo
    cosmo_nb = ccl.CosmologyVanillaLCDM(
        transfer_function='bbks', baryonic_effects=None)
    pk_nb = cosmo_nb.get_nonlin_power()
    pk_wb = bar.include_baryonic_effects(cosmo_nb, pk_nb)
    # 2. In cosmo - from object.
    cosmo_wb2 = ccl.CosmologyVanillaLCDM(
        transfer_function='bbks', baryonic_effects=bar)
    pk_wb2 = cosmo_wb2.get_nonlin_power()

    ks = np.geomspace(1E-2, 10, 128)
    pk_wb = pk_wb(ks, 1.0)
    pk_wb2 = pk_wb2(ks, 1.0)

    assert np.allclose(pk_wb, pk_wb2, atol=0, rtol=1E-6)


def test_baryons_in_cosmology_error():
    with pytest.raises(ValueError):
        ccl.CosmologyVanillaLCDM(baryonic_effects=3.1416)



__all__ = ("MassFuncJenkins01",)

import numpy as np

from . import MassFunc


class MassFuncJenkins01(MassFunc):
    """Implements the mass function of `Jenkins et al. 2001
    <https://arxiv.org/abs/astro-ph/0005260>`_. This parametrization
    is only valid for 'fof' masses.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
    """
    name = 'Jenkins01'

    def __init__(self, *,
                 mass_def="fof",
                 mass_def_strict=True):
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.Delta != "fof"

    def _setup(self):
        self.A = 0.315
        self.b = 0.61
        self.q = 3.8

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        return self.A * np.exp(-np.abs(-np.log(sigM) + self.b)**self.q)



__all__ = ("HaloProfileHOD",)

import numpy as np
from scipy.special import sici, erf

from . import HaloProfile


class HaloProfileHOD(HaloProfile):
    """ A generic halo occupation distribution (HOD)
    profile describing the number density of galaxies
    as a function of halo mass.

    The parametrization for the mean profile is:

    .. math::
       \\langle n_g(r)|M,a\\rangle = \\bar{N}_c(M,a)
       \\left[f_c(a)+\\bar{N}_s(M,a) u_{\\rm sat}(r|M,a)\\right]

    where :math:`\\bar{N}_c` and :math:`\\bar{N}_s` are the
    mean number of central and satellite galaxies respectively,
    :math:`f_c` is the observed fraction of central galaxies, and
    :math:`u_{\\rm sat}(r|M,a)` is the distribution of satellites
    as a function of distance to the halo centre.

    These quantities are parametrized as follows:

    .. math::
       \\bar{N}_c(M,a)=\\frac{1}{2}\\left[1+{\\rm erf}
       \\left(\\frac{\\log(M/M_{\\rm min})}{\\sigma_{{\\rm ln}M}}
       \\right)\\right]

    .. math::
       \\bar{N}_s(M,a)=\\Theta(M-M_0)\\left(\\frac{M-M_0}{M_1}
       \\right)^\\alpha

    .. math::
       u_s(r|M,a)\\propto\\frac{\\Theta(r_{\\rm max}-r)}
       {(r/r_g)(1+r/r_g)^2}

    Where :math:`\\Theta(x)` is the Heaviside step function,
    and the proportionality constant in the last equation is
    such that the volume integral of :math:`u_s` is 1. The
    radius :math:`r_g` is related to the NFW scale radius :math:`r_s`
    through :math:`r_g=\\beta_g\\,r_s`, and the radius
    :math:`r_{\\rm max}` is related to the overdensity radius
    :math:`r_\\Delta` as :math:`r_{\\rm max}=\\beta_{\\rm max}r_\\Delta`.
    The scale radius is related to the comoving overdensity halo
    radius through the concentration-mass relation via
    :math:`r_\\Delta(M) = c(M)\\,r_s`.

    All the quantities :math:`\\log_{10}M_{\\rm min}`,
    :math:`\\log_{10}M_0`, :math:`\\log_{10}M_1`,
    :math:`\\sigma_{{\\rm ln}M}`, :math:`f_c`, :math:`\\alpha`,
    :math:`\\beta_g` and :math:`\\beta_{\\rm max}` are
    time-dependent via a linear expansion around a pivot scale
    factor :math:`a_*` with an offset and a tilt parameter
    (:math:`X_0` and :math:`X_p`, respectively):

    .. math::
       X(a) = X_0 + X_p\\,(a-a_*).

    This definition of the HOD profile draws from several papers
    in the literature, including: `Zheng et al. 2005
    <https://arxiv.org/abs/astro-ph/0408564>`_, `Ando et al. 2018
    <https://arxiv.org/abs/1706.05422>`_, and `Nicola et al. 2020
    <https://arxiv.org/abs/1912.08209>`_. The default values used
    here are roughly compatible with those found in the latter
    paper.

    See :class:`~pyccl.halos.profiles_2pt.Profile2ptHOD` for a
    description of the Fourier-space two-point correlator of the
    HOD profile.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        concentration (:class:`~pyccl.halos.halo_model_base.Concentration`):
            concentration-mass relation to use with this profile.
        log10Mmin_0 (:obj:`float`): offset parameter for
            :math:`\\log_{10}M_{\\rm min}`.
        log10Mmin_p (:obj:`float`): tilt parameter for
            :math:`\\log_{10}M_{\\rm min}`.
        siglnM_0 (:obj:`float`): offset parameter for
            :math:`\\sigma_{{\\rm ln}M}`.
        siglnM_p (:obj:`float`): tilt parameter for
            :math:`\\sigma_{{\\rm ln}M}`.
        log10M0_0 (:obj:`float`): offset parameter for
            :math:`\\log_{10}M_0`.
        log10M0_p (:obj:`float`): tilt parameter for
            :math:`\\log_{10}M_0`.
        log10M1_0 (:obj:`float`): offset parameter for
            :math:`\\log_{10}M_1`.
        log10M1_p (:obj:`float`): tilt parameter for
            :math:`\\log_{10}M_1`.
        alpha_0 (:obj:`float`): offset parameter for
            :math:`\\alpha`.
        alpha_p (:obj:`float`): tilt parameter for
            :math:`\\alpha`.
        fc_0 (:obj:`float`): offset parameter for
            :math:`f_c`.
        fc_p (:obj:`float`): tilt parameter for
            :math:`f_c`.
        bg_0 (:obj:`float`): offset parameter for
            :math:`\\beta_g`.
        bg_p (:obj:`float`): tilt parameter for
            :math:`\\beta_g`.
        bmax_0 (:obj:`float`): offset parameter for
            :math:`\\beta_{\\rm max}`.
        bmax_p (:obj:`float`): tilt parameter for
            :math:`\\beta_{\\rm max}`.
        a_pivot (:obj:`float`): pivot scale factor :math:`a_*`.
        ns_independent (:obj:`bool`): drop requirement to only form
            satellites when centrals are present.
        is_number_counts (:obj:`bool`): set to ``True`` if this profile
            is meant to represent galaxy overdensity.
    """
    __repr_attrs__ = __eq_attrs__ = (
        "log10Mmin_0", "log10Mmin_p", "siglnM_0", "siglnM_p", "log10M0_0",
        "log10M0_p", "log10M1_0", "log10M1_p", "alpha_0", "alpha_p", "fc_0",
        "fc_p", "bg_0", "bg_p", "bmax_0", "bmax_p", "a_pivot",
        "_is_number_counts", "ns_independent", "mass_def", "concentration",
        "precision_fftlog",)

    def __init__(self, *, mass_def, concentration,
                 log10Mmin_0=12., log10Mmin_p=0., siglnM_0=0.4,
                 siglnM_p=0., log10M0_0=7., log10M0_p=0.,
                 log10M1_0=13.3, log10M1_p=0., alpha_0=1.,
                 alpha_p=0., fc_0=1., fc_p=0.,
                 bg_0=1., bg_p=0., bmax_0=1., bmax_p=0.,
                 a_pivot=1., ns_independent=False,
                 is_number_counts=True):
        self.log10Mmin_0 = log10Mmin_0
        self.log10Mmin_p = log10Mmin_p
        self.log10M0_0 = log10M0_0
        self.log10M0_p = log10M0_p
        self.log10M1_0 = log10M1_0
        self.log10M1_p = log10M1_p
        self.siglnM_0 = siglnM_0
        self.siglnM_p = siglnM_p
        self.alpha_0 = alpha_0
        self.alpha_p = alpha_p
        self.fc_0 = fc_0
        self.fc_p = fc_p
        self.bg_0 = bg_0
        self.bg_p = bg_p
        self.bmax_0 = bmax_0
        self.bmax_p = bmax_p
        self.a_pivot = a_pivot
        self.ns_independent = ns_independent
        super().__init__(mass_def=mass_def, concentration=concentration,
                         is_number_counts=is_number_counts)

    def update_parameters(self, *, log10Mmin_0=None, log10Mmin_p=None,
                          siglnM_0=None, siglnM_p=None,
                          log10M0_0=None, log10M0_p=None,
                          log10M1_0=None, log10M1_p=None,
                          alpha_0=None, alpha_p=None,
                          fc_0=None, fc_p=None,
                          bg_0=None, bg_p=None,
                          bmax_0=None, bmax_p=None,
                          a_pivot=None,
                          ns_independent=None):
        """ Update any of the parameters associated with
        this profile. Any parameter set to ``None`` won't be updated.

        Args:
            log10Mmin_0 (:obj:`float`): offset parameter for
                :math:`\\log_{10}M_{\\rm min}`.
            log10Mmin_p (:obj:`float`): tilt parameter for
                :math:`\\log_{10}M_{\\rm min}`.
            siglnM_0 (:obj:`float`): offset parameter for
                :math:`\\sigma_{{\\rm ln}M}`.
            siglnM_p (:obj:`float`): tilt parameter for
                :math:`\\sigma_{{\\rm ln}M}`.
            log10M0_0 (:obj:`float`): offset parameter for
                :math:`\\log_{10}M_0`.
            log10M0_p (:obj:`float`): tilt parameter for
                :math:`\\log_{10}M_0`.
            log10M1_0 (:obj:`float`): offset parameter for
                :math:`\\log_{10}M_1`.
            log10M1_p (:obj:`float`): tilt parameter for
                :math:`\\log_{10}M_1`.
            alpha_0 (:obj:`float`): offset parameter for
                :math:`\\alpha`.
            alpha_p (:obj:`float`): tilt parameter for
                :math:`\\alpha`.
            fc_0 (:obj:`float`): offset parameter for
                :math:`f_c`.
            fc_p (:obj:`float`): tilt parameter for
                :math:`f_c`.
            bg_0 (:obj:`float`): offset parameter for
                :math:`\\beta_g`.
            bg_p (:obj:`float`): tilt parameter for
                :math:`\\beta_g`.
            bmax_0 (:obj:`float`): offset parameter for
                :math:`\\beta_{\\rm max}`.
            bmax_p (:obj:`float`): tilt parameter for
                :math:`\\beta_{\\rm max}`.
            a_pivot (:obj:`float`): pivot scale factor :math:`a_*`.
            ns_independent (:obj:`bool`): drop requirement to only form
                satellites when centrals are present
        """
        if log10Mmin_0 is not None:
            self.log10Mmin_0 = log10Mmin_0
        if log10Mmin_p is not None:
            self.log10Mmin_p = log10Mmin_p
        if log10M0_0 is not None:
            self.log10M0_0 = log10M0_0
        if log10M0_p is not None:
            self.log10M0_p = log10M0_p
        if log10M1_0 is not None:
            self.log10M1_0 = log10M1_0
        if log10M1_p is not None:
            self.log10M1_p = log10M1_p
        if siglnM_0 is not None:
            self.siglnM_0 = siglnM_0
        if siglnM_p is not None:
            self.siglnM_p = siglnM_p
        if alpha_0 is not None:
            self.alpha_0 = alpha_0
        if alpha_p is not None:
            self.alpha_p = alpha_p
        if fc_0 is not None:
            self.fc_0 = fc_0
        if fc_p is not None:
            self.fc_p = fc_p
        if bg_0 is not None:
            self.bg_0 = bg_0
        if bg_p is not None:
            self.bg_p = bg_p
        if bmax_0 is not None:
            self.bmax_0 = bmax_0
        if bmax_p is not None:
            self.bmax_p = bmax_p
        if a_pivot is not None:
            self.a_pivot = a_pivot
        if ns_independent is not None:
            self.ns_independent = ns_independent

    def _usat_real(self, cosmo, r, M, a):
        r_use = np.atleast_1d(r)
        M_use = np.atleast_1d(M)

        # Comoving virial radius
        bg = self.bg_0 + self.bg_p * (a - self.a_pivot)
        bmax = self.bmax_0 + self.bmax_p * (a - self.a_pivot)
        R_M = self.mass_def.get_radius(cosmo, M_use, a) / a
        c_M = self.concentration(cosmo, M_use, a)
        R_s = R_M / c_M
        c_M *= bmax / bg

        x = r_use[None, :] / (R_s[:, None] * bg)
        prof = 1./(x * (1 + x)**2)
        # Truncate
        prof[r_use[None, :] > R_M[:, None]*bmax] = 0

        norm = 1. / (4 * np.pi * (bg*R_s)**3 * (np.log(1+c_M) - c_M/(1+c_M)))
        prof = prof[:, :] * norm[:, None]

        if np.ndim(r) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def _usat_fourier(self, cosmo, k, M, a):
        M_use = np.atleast_1d(M)
        k_use = np.atleast_1d(k)

        # Comoving virial radius
        bg = self.bg_0 + self.bg_p * (a - self.a_pivot)
        bmax = self.bmax_0 + self.bmax_p * (a - self.a_pivot)
        R_M = self.mass_def.get_radius(cosmo, M_use, a) / a
        c_M = self.concentration(cosmo, M_use, a)
        R_s = R_M / c_M
        c_M *= bmax / bg

        x = k_use[None, :] * R_s[:, None] * bg
        Si1, Ci1 = sici((1 + c_M[:, None]) * x)
        Si2, Ci2 = sici(x)

        P1 = 1. / (np.log(1+c_M) - c_M/(1+c_M))
        P2 = np.sin(x) * (Si1 - Si2) + np.cos(x) * (Ci1 - Ci2)
        P3 = np.sin(c_M[:, None] * x) / ((1 + c_M[:, None]) * x)
        prof = P1[:, None] * (P2 - P3)

        if np.ndim(k) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def _real(self, cosmo, r, M, a):
        r_use = np.atleast_1d(r)
        M_use = np.atleast_1d(M)

        Nc = self._Nc(M_use, a)
        Ns = self._Ns(M_use, a)
        fc = self._fc(a)
        # NFW profile
        ur = self._usat_real(cosmo, r_use, M_use, a)

        if self.ns_independent:
            prof = Nc[:, None] * fc + Ns[:, None] * ur
        else:
            prof = Nc[:, None] * (fc + Ns[:, None] * ur)

        if np.ndim(r) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def get_normalization(self, cosmo, a, *, hmc):
        """Returns the normalization of this profile, which is the
        mean galaxy number density.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology
                object.
            a (:obj:`float`): scale factor.
            hmc (:class:`~pyccl.halos.halo_model.HMCalculator`): a halo
                model calculator object.

        Returns:
            :obj:`float`: normalization factor of this profile.
        """
        def integ(M):
            Nc = self._Nc(M, a)
            Ns = self._Ns(M, a)
            fc = self._fc(a)
            if self.ns_independent:
                return Nc*fc + Ns
            return Nc*(fc + Ns)
        return hmc.integrate_over_massfunc(integ, cosmo, a)

    def _fourier(self, cosmo, k, M, a):
        M_use = np.atleast_1d(M)
        k_use = np.atleast_1d(k)

        Nc = self._Nc(M_use, a)
        Ns = self._Ns(M_use, a)
        fc = self._fc(a)
        # NFW profile
        uk = self._usat_fourier(cosmo, k_use, M_use, a)

        if self.ns_independent:
            prof = Nc[:, None] * fc + Ns[:, None] * uk
        else:
            prof = Nc[:, None] * (fc + Ns[:, None] * uk)

        if np.ndim(k) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def _fourier_variance(self, cosmo, k, M, a):
        # Fourier-space variance of the HOD profile
        M_use = np.atleast_1d(M)
        k_use = np.atleast_1d(k)

        Nc = self._Nc(M_use, a)
        Ns = self._Ns(M_use, a)
        fc = self._fc(a)
        # NFW profile
        uk = self._usat_fourier(cosmo, k_use, M_use, a)

        prof = Ns[:, None] * uk
        if self.ns_independent:
            prof = 2 * Nc[:, None] * fc * prof + prof**2
        else:
            prof = Nc[:, None] * (2 * fc * prof + prof**2)

        if np.ndim(k) == 0:
            prof = np.squeeze(prof, axis=-1)
        if np.ndim(M) == 0:
            prof = np.squeeze(prof, axis=0)
        return prof

    def _fc(self, a):
        # Observed fraction of centrals
        return self.fc_0 + self.fc_p * (a - self.a_pivot)

    def _Nc(self, M, a):
        # Number of centrals
        Mmin = 10.**(self.log10Mmin_0 + self.log10Mmin_p * (a - self.a_pivot))
        siglnM = self.siglnM_0 + self.siglnM_p * (a - self.a_pivot)
        return 0.5 * (1 + erf(np.log(M/Mmin)/siglnM))

    def _Ns(self, M, a):
        # Number of satellites
        M0 = 10.**(self.log10M0_0 + self.log10M0_p * (a - self.a_pivot))
        M1 = 10.**(self.log10M1_0 + self.log10M1_p * (a - self.a_pivot))
        alpha = self.alpha_0 + self.alpha_p * (a - self.a_pivot)
        return np.heaviside(M-M0, 1) * (np.fabs(M-M0) / M1)**alpha



__all__ = ("LagrangianPTCalculator",)

import numpy as np

from .. import (CCLAutoRepr, CCLError, Pk2D,
                get_pk_spline_a, unlock_instance)


# All valid Pk pair labels and their aliases
_PK_ALIAS = {
    'm:m': 'm:m', 'm:b1': 'm:b1', 'm:b2': 'm:b2',
    'm:b3nl': 'm:b3nl', 'm:bs': 'm:bs', 'm:bk2': 'm:bk2',
    'b1:b1': 'b1:b1', 'b1:b2': 'b1:b2', 'b1:b3nl': 'b1:b3nl',
    'b1:bs': 'b1:bs', 'b1:bk2': 'b1:bk2', 'b2:b2': 'b2:b2',
    'b2:b3nl': 'zero', 'b2:bs': 'b2:bs', 'b2:bk2': 'b2:bk2',
    'b3nl:b3nl': 'zero', 'b3nl:bs': 'zero',
    'b3nl:bk2': 'zero', 'bs:bs': 'bs:bs',
    'bs:bk2': 'bs:bk2', 'bk2:bk2': 'bk2:bk2'}


class LagrangianPTCalculator(CCLAutoRepr):
    """ This class implements a set of methods that can be
    used to compute the various components needed to estimate
    Lagrangian perturbation theory correlations. These calculations
    are currently based on velocileptors
    (https://github.com/sfschen/velocileptors).

    In the parametrisation used here, the galaxy overdensity
    is expanded as:

    .. math::
        \\delta_g=b_1\\,\\delta+\\frac{b_2}{2}\\delta^2+
        \\frac{b_s}{2}s^2+\\frac{b_{3nl}}{2}O_{3}+
        \\frac{b_{k2}}{2}\\nabla^2\\delta.

    where

    .. math::
        O_{3}(k) = s_{ij}(k)t_{ij}(k) +
        \\frac{16}{63}\\langle \\delta_{lin}\\rangle

    .. note:: As opposed to EPT, all terms including
              :math:`\\nabla^2\\delta`) are taken into account
              in the expansion.

    .. note:: Terms of the form
              :math:`\\langle \\delta^2 \\psi_{nl}\\rangle` (and
              likewise for :math:`s^2` and :math:`\\nabla^2\\delta`)
              are set to zero.

    .. note:: This calculator does not account for any form of
              stochastic bias contribution to the power spectra.
              If necessary, consider adding it in post-processing.

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
            If present, internal PT power spectrum templates will
            be initialized. If ``None``, you will need to initialize
            them using the :meth:`update_ingredients` method.
        log10k_min (:obj:`float`): decimal logarithm of the minimum
            Fourier scale (in :math:`{\\rm Mpc}^{-1}`) for which you want to
            calculate perturbation theory quantities.
        log10k_max (:obj:`float`): decimal logarithm of the maximum
            Fourier scale (in :math:`{\\rm Mpc}^{-1}`) for which you want to
            calculate perturbation theory quantities.
        nk_per_decade (:obj:`int` or :obj:`float`): number of k values per
            decade.
        a_arr (array): array of values of the scale factor at
            which all power spectra will be evaluated. If ``None``,
            the default sampling used internally by CCL will be
            used. Note that this may be slower than a bespoke sampling
            optimised for your particular application.
        k_cutoff (:obj:`float`): exponential cutoff scale. All power
            spectra will be multiplied by a cutoff factor of the
            form :math:`\\exp(-(k/k_*)^n)`, where :math:`k_*` is
            the cutoff scale. This may be useful when using the
            resulting power spectra to compute correlation
            functions if some of the PT contributions do not
            fall sufficiently fast on small scales. If ``None``
            (default), no cutoff factor will be applied.
        n_exp_cutoff (:obj:`float`): exponent of the cutoff factor (see
            ``k_cutoff``).
        b1_pk_kind (:obj:`str`): power spectrum to use for the first-order
            bias terms in the expansion. ``'linear'``: use the linear
            matter power spectrum. ``'nonlinear'``: use the non-linear
            matter power spectrum. ``'pt'``: use the 1-loop SPT matter
            power spectrum.
        bk2_pk_kind (:obj:`str`): power spectrum to use for the non-local
            bias terms in the expansion. Same options and default as
            ``b1_pk_kind``.
    """
    __repr_attrs__ = __eq_attrs__ = ('k_s', 'a_s', 'exp_cutoff',
                                     'b1_pk_kind', 'bk2_pk_kind')

    def __init__(self, *, cosmo=None,
                 log10k_min=-4, log10k_max=2, nk_per_decade=20,
                 a_arr=None, k_cutoff=None, n_exp_cutoff=4,
                 b1_pk_kind='nonlinear', bk2_pk_kind='nonlinear'):

        # k sampling
        nk_total = int((log10k_max - log10k_min) * nk_per_decade)
        self.k_s = np.logspace(log10k_min, log10k_max, nk_total)

        # a sampling
        if a_arr is None:
            a_arr = get_pk_spline_a()
        self.a_s = a_arr.copy()
        self.z_s = 1/self.a_s-1

        # Cutoff factor
        if k_cutoff is not None:
            self.exp_cutoff = np.exp(-(self.k_s/k_cutoff)**n_exp_cutoff)
            self.exp_cutoff = self.exp_cutoff[None, :]
        else:
            self.exp_cutoff = 1

        # b1/bk P(k) prescription
        if b1_pk_kind not in ['linear', 'nonlinear', 'pt']:
            raise ValueError(f"Unknown P(k) prescription {b1_pk_kind}")
        if bk2_pk_kind not in ['linear', 'nonlinear', 'pt']:
            raise ValueError(f"Unknown P(k) prescription {bk2_pk_kind}")
        self.b1_pk_kind = b1_pk_kind
        self.bk2_pk_kind = bk2_pk_kind

        # Initialize all expensive arrays to ``None``.
        self._cosmo = None

        # Fill them out if cosmo is present
        if cosmo is not None:
            self.update_ingredients(cosmo)

        # All valid Pk pair labels
        self._pk_valid = list(_PK_ALIAS.keys())
        # List of Pk2Ds to fill out
        self._pk2d_temp = {}

    def _check_init(self):
        if self.initialised:
            return
        raise CCLError("PT templates have not been initialised "
                       "for this calculator. Please do so using "
                       "`update_ingredients`.")

    @property
    def initialised(self):
        return hasattr(self, "pk_bk")

    @unlock_instance
    def update_ingredients(self, cosmo):
        """ Update the internal PT arrays.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        """
        if self.initialised and (cosmo == self._cosmo):
            return

        pklz0 = cosmo.linear_matter_power(self.k_s, 1.0)
        g = cosmo.growth_factor(self.a_s)

        from velocileptors.EPT.cleft_kexpanded_resummed_fftw import RKECLEFT
        h = cosmo['h']
        cleft = RKECLEFT(self.k_s / h, pklz0 * h ** 3)
        lpt_table = []
        for gz in g:
            cleft.make_ptable(D=gz, kmin=self.k_s[0] / h,
                              kmax=self.k_s[-1] / h, nk=self.k_s.size)
            lpt_table.append(cleft.pktable)
        lpt_table = np.array(lpt_table)
        lpt_table[:, :, 1:] /= h ** 3

        self.lpt_table = lpt_table
        self.one_loop_dd = self.lpt_table[:, :, 1]

        # b1/bk power spectrum
        pks = {}
        if 'nonlinear' in [self.b1_pk_kind, self.bk2_pk_kind]:
            pks['nonlinear'] = np.array([cosmo.nonlin_matter_power(self.k_s, a)
                                         for a in self.a_s])
        if 'linear' in [self.b1_pk_kind, self.bk2_pk_kind]:
            pks['linear'] = np.array([cosmo.linear_matter_power(self.k_s, a)
                                      for a in self.a_s])
        # If PT power spectrum is required it will be calculated on
        # the fly in the respective _get_pgg and _get_pgm functions
        # later as it needs the biases
        if 'pt' in [self.b1_pk_kind, self.bk2_pk_kind]:
            pks['pt'] = None
        self.pk_b1 = pks[self.b1_pk_kind]
        self.pk_bk = pks[self.bk2_pk_kind]

        # Reset template power spectra
        self._pk2d_temp = {}
        self._cosmo = cosmo

    def _get_pgg(self, tr1, tr2):
        """ Get the number counts auto-spectrum at the internal
        set of wavenumbers and scale factors.

        Args:
            tr1 (:class:`~pyccl.nl_pt.tracers.PTTracer`): first
                tracer to correlate.
            tr2 (:class:`~pyccl.nl_pt.tracers.PTTracer`): first
                tracer to correlate.

        Returns:
            array: 2D array of shape `(N_a, N_k)`, where `N_k` \
                is the size of this object's `k_s` attribute, and \
                `N_a` is the size of the object's `a_s` attribute.
        """
        self._check_init()
        # Get biases
        b11 = tr1.b1(self.z_s)
        b21 = tr1.b2(self.z_s)
        bs1 = tr1.bs(self.z_s)
        bk21 = tr1.bk2(self.z_s)
        b3nl1 = tr1.b3nl(self.z_s)
        b12 = tr2.b1(self.z_s)
        b22 = tr2.b2(self.z_s)
        bs2 = tr2.bs(self.z_s)
        bk22 = tr2.bk2(self.z_s)
        b3nl2 = tr2.b3nl(self.z_s)

        # Transform from Eulerian to Lagrangian biases
        bL11 = b11 - 1
        bL12 = b12 - 1
        # Get Pk templates
        if self.pk_b1 is None:
            Pdmdm = self.lpt_table[:, :, 1]
            Pdmd1 = 0.5*self.lpt_table[:, :, 2]
            Pd1d1 = self.lpt_table[:, :, 3]
            pgg = (Pdmdm + (bL11+bL12)[:, None] * Pdmd1 +
                   (bL11*bL12)[:, None] * Pd1d1)
        else:
            pgg = (b11*b12)[:, None]*self.pk_b1
        Pdmd2 = 0.5*self.lpt_table[:, :, 4]
        Pd1d2 = 0.5*self.lpt_table[:, :, 5]
        Pd2d2 = self.lpt_table[:, :, 6]
        Pdms2 = 0.25*self.lpt_table[:, :, 7]
        Pd1s2 = 0.25*self.lpt_table[:, :, 8]
        Pd2s2 = 0.25*self.lpt_table[:, :, 9]
        Ps2s2 = 0.25*self.lpt_table[:, :, 10]
        Pdmo3 = 0.25 * self.lpt_table[:, :, 11]
        Pd1o3 = 0.25 * self.lpt_table[:, :, 12]
        if self.pk_bk is not None:
            Pd1k2 = 0.5*self.pk_bk * (self.k_s**2)[None, :]
        else:
            Pdmdm = self.lpt_table[:, :, 1]
            Pdmd1 = 0.5*self.lpt_table[:, :, 2]

            Pdmk2 = 0.5*Pdmdm * (self.k_s**2)[None, :]
            Pd1k2 = 0.5*Pdmd1 * (self.k_s**2)[None, :]
            Pd2k2 = Pdmd2 * (self.k_s**2)[None, :]
            Ps2k2 = Pdms2 * (self.k_s**2)[None, :]
            Pk2k2 = 0.25*Pdmdm * (self.k_s**4)[None, :]

        pgg += ((b21 + b22)[:, None] * Pdmd2 +
                (bs1 + bs2)[:, None] * Pdms2 +
                (bL11*b22 + bL12*b21)[:, None] * Pd1d2 +
                (bL11*bs2 + bL12*bs1)[:, None] * Pd1s2 +
                (b21*b22)[:, None] * Pd2d2 +
                (b21*bs2 + b22*bs1)[:, None] * Pd2s2 +
                (bs1*bs2)[:, None] * Ps2s2 +
                (b3nl1 + b3nl2)[:, None] * Pdmo3 +
                (bL11*b3nl2 + bL12*b3nl1)[:, None] * Pd1o3)

        if self.pk_bk is not None:
            pgg += (b12*bk21+b11*bk22)[:, None] * Pd1k2
        else:
            pgg += ((bk21 + bk22)[:, None] * Pdmk2 +
                    (bL12 * bk21 + bL11 * bk22)[:, None] * Pd1k2 +
                    (b22 * bk21 + b21 * bk22)[:, None] * Pd2k2 +
                    (bs2 * bk21 + bs1 * bk22)[:, None] * Ps2k2 +
                    (bk21 * bk22)[:, None] * Pk2k2)

        return pgg*self.exp_cutoff

    def _get_pgm(self, trg):
        """ Get the number counts - matter cross-spectrum at the internal
        set of wavenumbers and scale factors.

        Args:
            trg (:class:`~pyccl.nl_pt.tracers.PTTracer`): number
                counts tracer.

        Returns:
            array: 2D array of shape `(N_a, N_k)`, where `N_k` \
                is the size of this object's `k_s` attribute, and \
                `N_a` is the size of the object's `a_s` attribute.
        """
        self._check_init()
        # Get biases
        b1 = trg.b1(self.z_s)
        b2 = trg.b2(self.z_s)
        bs = trg.bs(self.z_s)
        bk2 = trg.bk2(self.z_s)
        b3nl = trg.b3nl(self.z_s)

        # Compute Lagrangian bias
        bL1 = b1 - 1
        # Get Pk templates
        if self.pk_b1 is None:
            Pdmdm = self.lpt_table[:, :, 1]
            Pdmd1 = 0.5*self.lpt_table[:, :, 2]
            pgm = Pdmdm + bL1[:, None] * Pdmd1
        else:
            pgm = b1[:, None]*self.pk_b1
        Pdmd2 = 0.5*self.lpt_table[:, :, 4]
        Pdms2 = 0.25*self.lpt_table[:, :, 7]
        Pdmo3 = 0.25 * self.lpt_table[:, :, 11]

        if self.pk_bk is not None:
            Pdmk2 = 0.5*self.pk_bk * (self.k_s**2)[None, :]
        else:
            Pdmdm = self.lpt_table[:, :, 1]

            Pdmk2 = 0.5*Pdmdm * (self.k_s**2)[None, :]

        pgm += (b2[:, None] * Pdmd2 +
                bs[:, None] * Pdms2 +
                b3nl[:, None] * Pdmo3 +
                bk2[:, None] * Pdmk2)

        return pgm*self.exp_cutoff

    def _get_pmm(self):
        """ Get the one-loop matter power spectrum.

        Returns:
            array: 2D array of shape `(N_a, N_k)`, where `N_k` \
                is the size of this object's `k_s` attribute, and \
                `N_a` is the size of the object's `a_s` attribute.
        """
        self._check_init()

        pk = self.lpt_table[:, :, 1]

        return pk*self.exp_cutoff

    def get_biased_pk2d(self, tracer1, *, tracer2=None,
                        extrap_order_lok=1, extrap_order_hik=2):
        """Returns a :class:`~pyccl.pk2d.Pk2D` object containing
        the PT power spectrum for two quantities defined by
        two :class:`~pyccl.nl_pt.tracers.PTTracer` objects.

        Args:
            tracer1 (:class:`~pyccl.nl_pt.tracers.PTTracer`): the first
                tracer being correlated.
            tracer2 (:class:`~pyccl.nl_pt.tracers.PTTracer`): the second
                tracer being correlated. If ``None``, the auto-correlation
                of the first tracer will be returned.
            extrap_order_lok (:obj:`int`): extrapolation order to be used on
                k-values below the minimum of the splines. See
                :class:`~pyccl.pk2d.Pk2D`.
            extrap_order_hik (:obj:`int`): extrapolation order to be used on
                k-values above the maximum of the splines. See
                :class:`~pyccl.pk2d.Pk2D`.

        Returns:
            :class:`~pyccl.pk2d.Pk2D`: PT power spectrum.
        """
        if tracer2 is None:
            tracer2 = tracer1

        t1 = tracer1.type
        t2 = tracer2.type

        if t1 == 'IA' or t2 == 'IA':
            raise ValueError("Intrinsic alignments not implemented in "
                             "LagrangianPTCalculator.")

        if t1 == 'NC':
            if t2 == 'NC':
                pk = self._get_pgg(tracer1, tracer2)
            else:  # Must be matter
                pk = self._get_pgm(tracer1)
        else:  # Must be matter
            if t2 == 'NC':
                pk = self._get_pgm(tracer2)
            else:  # Must be matter
                pk = self._get_pmm()

        pk2d = Pk2D(a_arr=self.a_s,
                    lk_arr=np.log(self.k_s),
                    pk_arr=pk,
                    is_logp=False,
                    extrap_order_lok=extrap_order_lok,
                    extrap_order_hik=extrap_order_hik)
        return pk2d

    def get_pk2d_template(self, kind, *, extrap_order_lok=1,
                          extrap_order_hik=2):
        """Returns a :class:`~pyccl.pk2d.Pk2D` object containing
        the power spectrum template for two of the PT operators. The
        combination returned is determined by ``kind``, which must be
        a string of the form ``'q1:q2'``, where ``q1`` and ``q2`` denote
        the two operators whose power spectrum is sought. Valid
        operator names are: ``'m'`` (matter overdensity), ``'b1'``
        (first-order overdensity), ``'b2'`` (:math:`\\delta^2`
        term in galaxy bias expansion), ``'bs'`` (:math:`s^2` term
        in galaxy bias expansion), ``'b3nl'`` (:math:`\\psi_{nl}`
        term in galaxy bias expansion), ``'bk2'`` (non-local
        :math:`\\nabla^2 \\delta` term in galaxy bias expansion)

        Args:
            kind (:obj:`str`): string defining the pair of PT operators for
                which we want the power spectrum.
            extrap_order_lok (:obj:`int`): extrapolation order to be used on
                k-values below the minimum of the splines. See
                :class:`~pyccl.pk2d.Pk2D`.
            extrap_order_hik (:obj:`int`): extrapolation order to be used on
                k-values above the maximum of the splines. See
                :class:`~pyccl.pk2d.Pk2D`.

        Returns:
            :class:`~pyccl.pk2d.Pk2D`: PT power spectrum.
        """
        if not (kind in _PK_ALIAS):
            # Reverse order and check again
            kind_reverse = ':'.join(kind.split(':')[::-1])
            if not (kind_reverse in _PK_ALIAS):
                raise ValueError(f"Pk template {kind} not valid")
            kind = kind_reverse
        pk_name = _PK_ALIAS[kind]

        # If already built, return
        if pk_name in self._pk2d_temp:
            return self._pk2d_temp[pk_name]

        self._check_init()

        if pk_name == 'm:m':
            pk = self._get_pmm()
        elif pk_name == 'm:b1':
            pk = 0.5*self.lpt_table[:, :, 2]
        elif pk_name == 'm:b2':
            pk = 0.5*self.lpt_table[:, :, 4]
        elif pk_name == 'm:b3nl':
            pk = 0.25 * self.lpt_table[:, :, 11]
        elif pk_name == 'm:bs':
            pk = 0.25*self.lpt_table[:, :, 7]
        elif pk_name == 'm:bk2':
            Pdmdm = self.lpt_table[:, :, 1]
            pk = 0.5*Pdmdm * (self.k_s**2)[None, :]
        elif pk_name == 'b1:b1':
            pk = self.lpt_table[:, :, 3]
        elif pk_name == 'b1:b2':
            pk = 0.5*self.lpt_table[:, :, 5]
        elif pk_name == 'b1:b3nl':
            pk = 0.25 * self.lpt_table[:, :, 12]
        elif pk_name == 'b1:bs':
            pk = 0.25*self.lpt_table[:, :, 8]
        elif pk_name == 'b1:bk2':
            Pdmd1 = 0.5*self.lpt_table[:, :, 2]
            pk = 0.5*Pdmd1 * (self.k_s**2)[None, :]
        elif pk_name == 'b2:b2':
            pk = self.lpt_table[:, :, 6]
        elif pk_name == 'b2:bs':
            pk = 0.25*self.lpt_table[:, :, 9]
        elif pk_name == 'b2:bk2':
            Pdmd2 = 0.25*self.lpt_table[:, :, 4]
            pk = Pdmd2 * (self.k_s**2)[None, :]
        elif pk_name == 'bs:bs':
            pk = 0.25*self.lpt_table[:, :, 10]
        elif pk_name == 'bs:bk2':
            Pdms2 = 0.25*self.lpt_table[:, :, 7]
            pk = Pdms2 * (self.k_s**2)[None, :]
        elif pk_name == 'bk2:bk2':
            Pdmdm = self.lpt_table[:, :, 1]
            pk = 0.25*Pdmdm * (self.k_s**4)[None, :]
        elif pk_name == 'zero':
            # If zero, store None and return
            self._pk2d_temp[pk_name] = None
            return None

        # Build interpolator
        pk2d = Pk2D(a_arr=self.a_s,
                    lk_arr=np.log(self.k_s),
                    pk_arr=pk,
                    is_logp=False,
                    extrap_order_lok=extrap_order_lok,
                    extrap_order_hik=extrap_order_hik)

        # Store and return
        self._pk2d_temp[pk_name] = pk2d
        return pk2d



import numpy as np
import pytest
import pyccl as ccl

HALOPROFILE_TOLERANCE = 1E-3

COSMO = ccl.Cosmology(
    Omega_b=0.0486,
    Omega_c=0.2603,
    h=0.6774,
    n_s=0.9667,
    sigma8=0.8159,
    w0=-1,
    wa=0,
    m_nu=0,
    mass_split='normal',
    Neff=3.046, T_CMB=2.725,
    Omega_k=0,
    transfer_function='eisenstein_hu')


def test_profile_Hernquist():
    data = np.loadtxt("./benchmarks/data/haloprofile_hernquist_colossus.txt")
    a = 1.0
    halomass = 6e13
    concentration = 5
    mDelta = 200
    rmin = 0.01
    rmax = 100
    r = np.exp(
        np.log(rmin) +
        np.log(rmax/rmin) * np.arange(data.shape[0]) / (data.shape[0]-1))

    mdef = ccl.halos.MassDef(mDelta, 'matter')
    c = ccl.halos.ConcentrationConstant(c=concentration, mass_def=mdef)
    p = ccl.halos.HaloProfileHernquist(mass_def=mdef, concentration=c,
                                       truncated=False)

    prof = p.real(COSMO, r, halomass, a)

    tol = np.clip(np.abs(HALOPROFILE_TOLERANCE * data[:, 1]), 1e-12, np.inf)
    err = np.abs(prof - data[:, 1])
    assert np.all(err <= tol)


def test_profile_Einasto():
    data = np.loadtxt("./benchmarks/data/haloprofile_einasto_colossus.txt")
    a = 1.0
    halomass = 6e13
    concentration = 5
    mDelta = 200
    rmin = 0.01
    rmax = 100
    r = np.exp(
        np.log(rmin) +
        np.log(rmax/rmin) * np.arange(data.shape[0]) / (data.shape[0]-1))

    mdef = ccl.halos.MassDef(mDelta, 'matter')
    c = ccl.halos.ConcentrationConstant(c=concentration, mass_def=mdef)
    p = ccl.halos.HaloProfileEinasto(mass_def=mdef, concentration=c,
                                     truncated=False)

    prof = p.real(COSMO, r, halomass, a)

    tol = np.clip(np.abs(HALOPROFILE_TOLERANCE * data[:, 1]), 1e-12, np.inf)
    err = np.abs(prof - data[:, 1])
    assert np.all(err <= tol)


def test_profile_NFW():
    data = np.loadtxt("./benchmarks/data/haloprofile_nfw_colossus.txt")
    a = 1.0
    halomass = 6e13
    concentration = 5
    mDelta = 200
    rmin = 0.01
    rmax = 100
    r = np.exp(
        np.log(rmin) +
        np.log(rmax/rmin) * np.arange(data.shape[0]) / (data.shape[0]-1))

    mdef = ccl.halos.MassDef(mDelta, 'matter')
    c = ccl.halos.ConcentrationConstant(c=concentration, mass_def=mdef)
    p = ccl.halos.HaloProfileNFW(mass_def=mdef, concentration=c,
                                 truncated=False)

    prof = p.real(COSMO, r, halomass, a)

    tol = np.clip(np.abs(HALOPROFILE_TOLERANCE * data[:, 1]), 1e-12, np.inf)
    err = np.abs(prof - data[:, 1])
    assert np.all(err <= tol)


@pytest.mark.parametrize(
    'model', ['nfw', 'projected_nfw', 'einasto', 'hernquist'])
def test_haloprofile(model):

    data = np.loadtxt("./benchmarks/data/haloprofile_%s_colossus.txt" % model)
    a = 1.0
    concentration = 5
    halomass = 6e13
    halomassdef = 200
    rmin = 0.01
    rmax = 100
    r = np.exp(
        np.log(rmin) +
        np.log(rmax/rmin) * np.arange(data.shape[0]) / (data.shape[0]-1))

    mdef = ccl.halos.MassDef(halomassdef, 'matter')
    c = ccl.halos.ConcentrationConstant(c=concentration, mass_def=mdef)

    if model == 'nfw':
        p = ccl.halos.HaloProfileNFW(mass_def=mdef, concentration=c,
                                     truncated=False)
        prof = p.real(COSMO, r, halomass, a)
    elif model == 'projected_nfw':
        p = ccl.halos.HaloProfileNFW(mass_def=mdef, concentration=c,
                                     truncated=False, projected_analytic=True)
        prof = p.projected(COSMO, r, halomass, a)
    elif model == 'einasto':
        mdef = ccl.halos.MassDef(halomassdef, 'matter')
        p = ccl.halos.HaloProfileEinasto(mass_def=mdef, concentration=c,
                                         truncated=False)
        prof = p.real(COSMO, r, halomass, a)
    elif model == 'hernquist':
        p = ccl.halos.HaloProfileHernquist(mass_def=mdef, concentration=c,
                                           truncated=False)
        prof = p.real(COSMO, r, halomass, a)

    tol = np.clip(np.abs(HALOPROFILE_TOLERANCE * data[:, 1]), 1e-12, np.inf)
    err = np.abs(prof - data[:, 1])
    assert np.all(err <= tol)


def test_weak_lensing_functions():
    data = np.loadtxt("./benchmarks/data/haloprofile_nfw_wl_numcosmo.txt")
    z_lens = 1.0
    z_source = 2.0
    a_lens = 1.0 / (1.0 + z_lens)
    a_source = 1.0 / (1.0 + z_source)
    halomass = 1.0e15
    concentration = 5
    mDelta = "vir"  # 200
    rmin = 0.01
    rmax = 100
    r = np.exp(
        np.log(rmin) +
        np.log(rmax/rmin) * np.arange(data.shape[0]) / (data.shape[0]-1))

    r_al = r / a_lens
    len_r = len(r)
    a_source = [a_source]*len_r

    mdef = ccl.halos.MassDef(mDelta, 'matter')
    c = ccl.halos.ConcentrationConstant(c=concentration, mass_def=mdef)
    p = ccl.halos.HaloProfileNFW(mass_def=mdef, concentration=c,
                                 truncated=False, projected_analytic=True,
                                 cumul2d_analytic=True)

    kappa = p.convergence(COSMO, r_al, halomass,
                          a_lens=a_lens, a_source=a_source)
    tol = np.clip(np.abs(HALOPROFILE_TOLERANCE * data[:, 1]), 1e-12, np.inf)
    err_kappa = np.abs(kappa - data[:, 1])
    assert np.all(err_kappa <= tol)

    gamma = p.shear(COSMO, r_al, halomass,
                    a_lens=a_lens, a_source=a_source)
    tol = np.clip(np.abs(HALOPROFILE_TOLERANCE * data[:, 2]), 1e-12, np.inf)
    err_gamma = np.abs(gamma - data[:, 2])
    assert np.all(err_gamma <= tol)

    gt = p.reduced_shear(COSMO, r_al, halomass,
                         a_lens=a_lens, a_source=a_source)
    tol = np.clip(np.abs(HALOPROFILE_TOLERANCE * data[:, 3]), 1e-12, np.inf)
    err_gt = np.abs(gt - data[:, 3])
    assert np.all(err_gt <= tol)

    mu = p.magnification(COSMO, r_al, halomass,
                         a_lens=a_lens, a_source=a_source)
    tol = np.clip(np.abs(HALOPROFILE_TOLERANCE * data[:, 4]), 1e-12, np.inf)
    err_mu = np.abs(mu - data[:, 4])
    assert np.all(err_mu <= tol)



import pysz_gal.pysz_gal as szgal
import numpy as np
from scipy.integrate import simps


# Set N(z)
def _nz_2mrs(z):
    # From 1706.05422
    m = 1.31
    beta = 1.64
    x = z / 0.0266
    return x**m * np.exp(-x**beta)
z1 = 1e-5
z2 = 0.1
z_arr = np.linspace(z1, z2, 10024)
dndz = _nz_2mrs(z_arr)
dndz /= simps(dndz, z_arr)

# Set things up
pyszgal_cl = szgal.tsz_gal_cl()
ell = np.unique(np.geomspace(2, 1000, 20).astype(int)).astype(float)

# Set params
h0 = 0.67
Omb = 0.05
Omc = 0.25
As = 2E-9
ns = 0.9645
mnu = 0.00001
ombh2 = Omb*h0**2
omch2 = Omc*h0**2
bias = 1.55
lMcut = 11.8
Mcut = 10**lMcut
lM1 = 11.73
M1 = 10**lM1
kappa = 1.0
sigma_Ncen = 0.15
alp_Nsat = 0.77
rmax = 4.39
rgs = 1.17
pars = {'h0':h0, 'obh2':ombh2,'och2':omch2,\
        'As':As,'ns':ns,'mnu':mnu,\
        'mass_bias':bias,\
        'Mcut':Mcut,'M1':M1,'kappa':kappa,'sigma_Ncen':sigma_Ncen,'alp_Nsat':alp_Nsat,\
        'rmax':rmax,'rgs':rgs,\
        'flag_nu':True,'flag_tll':False}

# Run pysz_gal
gg, gy, tll, ng, tmp, ks, zs, pks = pyszgal_cl.get_tsz_cl(ell, pars, dndz,\
                                                          z1, z2, z1, z2, 101, 1001)
# Save benchmark data
np.savetxt("../k_hod.txt", ks)
np.savetxt("../z_hod.txt", zs)
np.savetxt("../pk_hod.txt", pks)
np.savetxt("../cl_hod.txt", np.transpose([ell, gg[0]+gg[1]]))



import numpy as np
import matplotlib.pyplot as plt
import py_cosmo_mad as csm

#Produces all sigma(M) benchmarks
#Contact david.alonso@physics.ox.ac.uk if you have issues running this script

TCMB=2.725
PLOT_STUFF=0
WRITE_STUFF=1
FS=16
LKMAX=7

def do_all(m_arr,cpar,prefix) :
    pcs=csm.PcsPar()
    pcs.background_set(cpar['om'],cpar['ol'],cpar['ob'],cpar['w0'],cpar['wa'],cpar['hh'],TCMB)
    pcs.set_linear_pk('BBKS',-3,LKMAX,0.01,cpar['ns'],cpar['s8'])

    r_arr=np.array([pcs.M2R(m) for m in m_arr])
    sm_arr=np.sqrt(np.array([pcs.sig0_L(r,r,'TopHat','TopHat') for r in r_arr]))

    if PLOT_STUFF==1 :
        plt.plot(m_arr,sm_arr)
        plt.xlabel('$M\\,[M_{\\odot}\\,h^{-1}]$',fontsize=FS)
        plt.ylabel('$\\sigma(M)$',fontsize=FS)
        plt.gca().set_xscale('log');
        plt.gca().set_yscale('log');
        plt.show()

    if WRITE_STUFF==1 :
        np.savetxt(prefix+"_sm.txt",np.transpose([m_arr,sm_arr]),header="[1] M (M_sun/h), [2] sigma(M)")

z_arr=np.array([0.,1.,2.,3.,4.,5.])
lm_arr=6.+2*np.arange(6)
m_arr=10**lm_arr

cpar_model1={'om': 0.3,'ol': 0.7,'ob':0.05,'hh': 0.7,'s8': 0.8,'ns': 0.96,'w0': -1.0, 'wa': 0.0}
cpar_model2={'om': 0.3,'ol': 0.7,'ob':0.05,'hh': 0.7,'s8': 0.8,'ns': 0.96,'w0': -0.9, 'wa': 0.0}
cpar_model3={'om': 0.3,'ol': 0.7,'ob':0.05,'hh': 0.7,'s8': 0.8,'ns': 0.96,'w0': -0.9, 'wa': 0.1}

do_all(m_arr,cpar_model1,"model1")
do_all(m_arr,cpar_model2,"model2")
do_all(m_arr,cpar_model3,"model3")
